/*
Copyright (c) 1991, 1992, 1993, 1994 Xerox Corporation.  All Rights Reserved.

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: iluPrmodule.c,v 1.53 1995/05/04 02:17:58 janssen Exp $
*/

#include <stdio.h>
#include <sys/types.h>	/* for gid_t */

/* from ILU */
#include <iluxport.h>
#include <iluaunix.h>	/* for ilu_AS_SunRPCUnix_UserID */
/* local */
#include "python.h"
#include "ilualobject.h"
#include "ilucaobject.h"
#include "iluclobject.h"
#include "iluftobject.h"
#include "ilulrobject.h"
#include "ilusvobject.h"
#include "iohcobject.h"
#include "ivobject.h"
#include "thcobject.h"

/* from Python */
#include "sysmodule.h"


#define Py_COUNT(x)	(((PyObject *)(x))->ob_refcnt)

#if (PYTHON_API_VERSION >= 1002)
#define PYTHON_LONG_FROM_STRING(str,base)	PyLong_FromString (str, ILU_NIL, base)
#else
#define PYTHON_LONG_FROM_STRING(str,base)	PyLong_FromString (str, base)
#endif /* PYTHON_API_VERSION >= 1002 */

/* exceptions defined by this module */
static PyObject *		GeneralError;
static PyObject *		ProtocolError;
static PyObject *		UnimplementedMethodError;

/* strings used in exceptions */
static char strArg1ShouldBeIluCall[] = "arg1 should be ilu_Call";
static char strArg2ShouldBeIntOrLongInt[] = "arg2 should be int or long int";
static char strArg2OutOfRange[] = "arg2 is out of range";
static char strListTooLong[] = "list length exceeds limit";
static char strListDoesntMatch[] = "list length doesn't match array bound";

/* time interval used in GC cleanups */
const ilu_FineTime plus1Minute = { 60, 0 };

/* names */
static char nameVarClass[]		= "_IluClass";
static char nameVarInstVars[]		= "_IluInstVars";
static char nameVarInstHandle[]		= "IluInstHandle";
static char nameVarServer[]		= "IluServer";

/* names used in Sun RPC authentication */
static char nameSunRPCAuth[]		= "sunrpc-unix";
static char nameSunRPCAuthUID[]		= "uid";
static char nameSunRPCAuthGID[]		= "gid";
static char nameSunRPCAuthHostname[]	= "hostname";
static char nameSunRPCAuthGroups[]	= "groups";

/* global state */
static PyObject *	defaultTrueServer;
static int		stopMainLoop;
static ilu_refany	gcAlarm = ILU_NIL;
static ilu_boolean	gcAlarmSet = ilu_FALSE;
static PyObject *	gcList = ILU_NIL;

/********************************/

static char *
stringDup(char *src)
{
	char *	result;

	if ((result = PyMem_NEW(char, strlen(src) + 1)) == 0)
	{
		(void) PyErr_NoMemory();
		return 0;
	}
	strcpy(result, src);
	return result;
}

static void
handleCalloutException(char *culprit)
{
	PyObject *	except;
	PyObject *	val;
	PyObject *	f;

#if (PYTHON_API_VERSION >= 1002)

	PyObject *	traceback;

	PyErr_Fetch (&except, &val, &traceback);
	Py_XDECREF (traceback);
	PyErr_Clear ();

#else

	PyErr_GetAndClear(&except, &val);

#endif /* PYTHON_API_VERSION >= 1002 */

	if (except == PyExc_SystemExit)
	{
		if (val == 0 || val == Py_None)
			Py_Exit(0);
		if (PyInt_Check(val))
			Py_Exit(PyInt_AsLong(val));
	}
#if (PYTHON_API_VERSION >= 1002)

	if ((f = PySys_GetObject("stderr")) != 0)
	  {
	    PyObject *trace = PyTraceBack_Fetch();
	    if (trace != 0)
	      {
		PyTraceBack_Print (trace, f);
		Py_DECREF (trace);
	      }
	  }

#else

	if ((f = PySys_GetObject("stderr")) != 0)
		PyErr_PrintTraceBack(f);

#endif /* PYTHON_API_VERSION >= 1002 */

	fprintf(stderr, "%s raised exception ", culprit);
	PyObject_Print(except, stderr, Py_PRINT_RAW);
	if (val && val != Py_None)
	{
		fprintf(stderr, ": ");
		PyObject_Print(val, stderr, Py_PRINT_RAW);
	}
	fprintf(stderr, "\n");
	PyErr_Clear();
}

/********************************/

static PyObject *
ilumod_SetDebugLevel(PyObject *self, PyObject *args)
{
	long	bits;

	if (!PyArg_Parse(args, "l", &bits))
		return 0;
	ilu_SetDebugLevel(bits);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_SetDebugLevelViaString(PyObject *self, PyObject *args)
{
	char *	spec;

	if (!PyArg_Parse(args, "z", &spec))
		return 0;
	ilu_SetDebugLevelViaString(spec);
	Py_INCREF(Py_None);
	return Py_None;
}

/********************************/

static ilu_Class
getKernelClass(PyClassObject *pycl)
{
	PyObject *	cl;
	char		msg[256];

	if ((cl = PyObject_GetAttrString((PyObject *) pycl, nameVarClass)) == 0)
	{
		sprintf(msg, "class arg should have a variable named %s",
			nameVarClass);
		PyErr_SetString(PyExc_TypeError, msg);
		return 0;
	}
	if (!ilucl_Check(cl))
	{
		sprintf(msg, "class arg's %s variable should be ilu_Class",
			nameVarClass);
		PyErr_SetString(PyExc_TypeError, msg);
		return 0;
	}
	return &((IluclObject *) cl)->c;
}

/********************************/

typedef struct CrNode CrNode;
struct CrNode
{
	PyClassObject *	pycl;
	ilu_Class	kclass;
	CrNode *	next;
};

static CrNode *	crHead;

static int
addToClassRegistry(PyClassObject *pycl)
{
	CrNode *	n	= PyMem_NEW(CrNode, 1);
	ilu_Class	kclass;

	if ((kclass = getKernelClass(pycl)) == 0)
		return -1;
	if (n == 0)
	{
		(void) PyErr_NoMemory();
		return -1;
	}
	Py_INCREF(pycl);
	n->pycl = pycl;
	n->kclass = kclass;
	n->next = crHead;
	crHead = n;
	return 0;
}

static PyObject *
findInClassRegistry(ilu_Class kclass)
{
	CrNode *	n;

	for (n = crHead; n != 0; n = n->next)
	{
		if (n->kclass == kclass)
			return (PyObject *) n->pycl;
	}
	return 0;
}

/********************************/

static PyObject *
ilumod_FormClassRecord(PyObject *self, PyObject *args)
{
	char *		name;
	char *		brand;
	char *		uniqueId;
	char *		singleton;
	char		collectible;
	char		optional;
	char *		authentication;
	PyObject *	methodTuple;
	PyObject *	superclassTuple;
	PyObject *	result;

	if (!PyArg_Parse(args, "(ssszbbzOO)", &name, &brand, &uniqueId,
	    &singleton, &collectible, &optional, &authentication,
	    &methodTuple, &superclassTuple))
		return 0;
	if (!PyTuple_Check(methodTuple))
	{
		PyErr_SetString(PyExc_TypeError, "arg7 should be tuple");
		return 0;
	}
	if (!PyTuple_Check(superclassTuple))
	{
		PyErr_SetString(PyExc_TypeError, "arg8 should be tuple");
		return 0;
	}
	result = ilucl_New(name, brand, uniqueId, singleton, collectible,
		optional, authentication, methodTuple, superclassTuple);
	return result;
}

/* Forward declarations. */
static int		ensureGcClient(void);

static PyObject *
ilumod_RegisterClass(PyObject *self, PyObject *args)
{
	PyClassObject *	pycl;
	ilu_Class	kclass;

	if (!PyArg_Parse(args, "O", &pycl))
		return 0;
	if (!PyClass_Check(pycl))
	{
		PyErr_SetString(PyExc_TypeError, "arg1 should be class");
		return 0;
	}
	if ((kclass = getKernelClass(pycl)) == 0)
		return 0;
	if (addToClassRegistry(pycl) < 0)
		return 0;
	ilu_RegisterClass(kclass);
	if (kclass->cl_collectible && ensureGcClient() < 0)
		return 0;
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_RegisterSkeletons(PyObject *self, PyObject *args)
{
	IluclObject *	cl;
	PyObject *	skelTuple;

	if (!PyArg_Parse(args, "(OO)", &cl, &skelTuple))
		return 0;
	if (!ilucl_Check(cl))
	{
		PyErr_SetString(PyExc_TypeError, "arg1 should be ilu_Class");
		return 0;
	}
	if (!PyTuple_Check(skelTuple))
	{
		PyErr_SetString(PyExc_TypeError, "arg2 should be tuple");
		return 0;
	}
	if (ilucl_RegisterSkeletons(cl, skelTuple) < 0)
		return 0;
	Py_INCREF(Py_None);
	return Py_None;
}

/********************************/

#define NullClObjectPtr			((IluclObject *) 0)

#define KClassMemberOffset		\
	((char *) &NullClObjectPtr->c - (char *) NullClObjectPtr)

#define KClassMemberToClObject(kclass)	\
	((IluclObject *) ((char *) kclass - KClassMemberOffset))

static void
callSkeleton(ilu_Call call, ilu_Class kclass, ilu_Method meth)
{
	IluclObject *	cl	= KClassMemberToClObject(kclass);
	int		methodIndex;
	PyObject *	skelFunc;
	PyObject *	skelArgs;
	PyObject *	ca;
	PyObject *	result;

	if (!ilucl_Check(cl))
	{
		fprintf(stderr, "callSkeleton: invalid kclass\n");
		return;
	}

	methodIndex = meth - cl->c.cl_methods;
	if (methodIndex < 0 || cl->c.cl_method_count <= methodIndex)
	{
		fprintf(stderr, "callSkeleton: bad method index (%d)\n",
			methodIndex);
		return;
	}
	if (cl->skeletons == 0 || (skelFunc = cl->skeletons[methodIndex]) == 0)
	{
		fprintf(stderr,
			"callSkeleton: no skeleton for method index (%d)\n",
			methodIndex);
		return;
	}
	if (!PyCallable_Check(skelFunc))
	{
		fprintf(stderr,
			"callSkeleton: bad skeleton for method index (%d)\n",
			methodIndex);
		return;
	}

	if ((ca = iluca_FromCall(call)) == 0)
	{
		fprintf(stderr, "callSkeleton: couldn't create ilu_Call\n");
		return;
	}
	if ((skelArgs = Py_BuildValue("(O)", ca)) == 0)
	{
		fprintf(stderr, "callSkeleton: couldn't build args\n");
		Py_DECREF(ca);
		return;
	}

	result = PyEval_CallObject(skelFunc, skelArgs);
	Py_DECREF(skelArgs);

	Py_DECREF(ca);

	if (result == 0)
		handleCalloutException("skeleton");
	else
		Py_DECREF(result);
}

static void	readServiceRequest(int fd, ilu_private rock);

static void
enableRequestsOnConn(ilu_Connection conn)
{
	int	fd	= ilu_FileDescriptorOfConnection(conn);

	if (ilu_RegisterInputSource(fd, readServiceRequest, conn) == 0)
	{
		fprintf(stderr, "couldn't register fd %d as input source\n",
			fd);
		return;
	}
}

static void
disableRequestsOnConn(ilu_Connection conn)
{
	int	fd	= ilu_FileDescriptorOfConnection(conn);

	if (ilu_UnregisterInputSource(fd) == 0)
	{
		fprintf(stderr, "couldn't unregister fd %d as input source\n",
			fd);
		return;
	}
}

static void
enableRequestsOnConnIfNotConcurrent(ilu_Connection conn)
{
	if (ilu_ThreadPerRequest(conn) == 0)
		enableRequestsOnConn(conn);
}

static void
enableRequestsOnConnIfConcurrent(ilu_Connection conn)
{
	if (ilu_ThreadPerRequest(conn))
		enableRequestsOnConn(conn);
}

static void
disableRequestsOnConnIfConnConcurrent(ilu_Connection conn)
{
	if (ilu_ThreadPerRequest(conn))
		disableRequestsOnConn(conn);
}

/* before: conn registered
**         (meaning readServiceRequest is registered for conn's fd)
** after: conn registered iff conn not closed
*/
static void
readServiceRequest(int fd, ilu_private rock)
{
	ilu_Connection			conn	= (ilu_Connection) rock;
	ilu_ReceiveRequestStatus	stat;
	ilu_Call			call;
	ilu_Class			kclass;
	ilu_Method			meth;
	ilu_cardinal			serialNo;

	if (fd != ilu_FileDescriptorOfConnection(conn))
	{
		fprintf(stderr, "fd (%d) != fd of connection (%d)\n",
			fd, ilu_FileDescriptorOfConnection(conn));
		return;
	}

	disableRequestsOnConn(conn);
	stat = ilu_ReceiveRequest(conn, &call, &kclass, &meth, &serialNo);
	switch (stat)
	{
	case ilu_conn_closed:
		ilu_CloseConnection(conn);
		break;

	case ilu_no_request:
	case ilu_builtin_meth:
	case ilu_read_failed:
	case ilu_not_request:
	case ilu_interp_failed:
		enableRequestsOnConn(conn);
		break;

	case ilu_good_request:
		callSkeleton(call, kclass, meth);
		/* skeleton leaves conn registered */
		break;

	default:
		fprintf(stderr, "bad request status (%d)\n", stat);
		enableRequestsOnConn(conn);
		break;
	}
}

/* after: new connection registered */
static void
readConnectionRequest(int fd, ilu_private rock)
{
	ilu_Port	port	= (ilu_Port) rock;
	ilu_Connection	conn;
	ilu_boolean	closed;

	if ((conn = ilu_HandleNewConnection(port, &closed)) == 0)
	{
		fprintf(stderr, "ilu_HandleNewConnection failed\n");
		return;
	}
	enableRequestsOnConn(conn);
}

static ilu_Port
createPort(ilu_Server kserver, char *transport, char *protocol)
{
	ilu_Port	port;
	int		fd;
	ilu_boolean	closed;

	if ((port = ilu_CreatePort(kserver, protocol, transport)) == 0)
		return 0;
	fd = ilu_FileDescriptorOfMooringOfPort(port, &closed);
	if (closed)
	{
		PyErr_SetString(GeneralError, "port is closed");
		return 0;
	}
	if (ilu_RegisterInputSource(fd, readConnectionRequest, port) == 0)
	{
		PyErr_SetString(GeneralError, "register input source failed");
		return 0;
	}
	return port;
}

static ilu_Server
createTrueServer(char *serverId, ilu_ObjectTable objTab, char *transport,
	char *protocol)
{
	ilu_Server	kserver;

	if ((kserver = ilu_CreateTrueServer(serverId, objTab)) == 0)
	{
		PyErr_SetString(GeneralError, "unable to create true server");
		return 0;
	}
	if (createPort(kserver, transport, protocol) == 0)
		return 0;
	return kserver;
}

static char *
getDefaultTransport(void)
{
	return "tcp_localhost_0";
}

static char *
getDefaultProtocol(const char *transport)
{
	return "sunrpc_";
}

/* Forward declarations. */
static ilu_ObjectTable	createObjectTable(PyObject *objectOfIh);
static void		setObjectTableServer(ilu_ObjectTable self,
				ilu_Server kserver);

static PyObject *
ilumod_CreateServer(PyObject *self, PyObject *args)
{
	char *		serverId;
	ilu_ObjectTable	objTab;
	char *		trans;
	char *		proto;
	PyObject *	objectOfIh;
	ilu_Server	kserver;
	PyObject *	result;

	if (PyArg_Parse(args, ""))
	{
		serverId = 0;
		trans = 0;
		proto = 0;
		objectOfIh = Py_None;
	}
	else if (PyErr_Clear(), PyArg_Parse(args, "z", &serverId))
	{
		trans = 0;
		proto = 0;
		objectOfIh = Py_None;
	}
	else if (PyErr_Clear(), PyArg_Parse(args, "(zz)", &serverId, &trans))
	{
		proto = 0;
		objectOfIh = Py_None;
	}
	else if (PyErr_Clear(), PyArg_Parse(args, "(zzz)", &serverId, &trans,
	    &proto))
	{
		objectOfIh = Py_None;
	}
	else
	{
		PyErr_Clear();
		if (!PyArg_Parse(args, "(zzzO)", &serverId, &trans, &proto,
		    &objectOfIh))
			return 0;
	}
	if (objectOfIh == Py_None)
		objTab = 0;
	else
	{
		if (!PyCallable_Check(objectOfIh))
		{
			PyErr_SetString(PyExc_TypeError,
				"arg4 should be callable");
			return 0;
		}
		if ((objTab = createObjectTable(objectOfIh)) == 0)
			return 0;
	}

	if (serverId == 0)
		serverId = ilu_InventID();
	else if ((serverId = stringDup(serverId)) == 0)
		return 0;
	if (trans == 0)
		trans = getDefaultTransport();
	if (proto == 0)
		proto = getDefaultProtocol(trans);

	if ((kserver = createTrueServer(serverId, objTab, trans, proto)) == 0)
	{
		if (objTab)
			(*objTab->ot_free_self)(objTab);
		return 0;
	}
	if (objTab)
		setObjectTableServer(objTab, kserver);
	if ((result = ilusv_FromServer(kserver)) == 0)
		ilu_BankServer(kserver);
	else if (defaultTrueServer == 0)
	{
		defaultTrueServer = result;
		Py_INCREF(result);
	}
	return result;
}

static PyObject *
createDefaultTrueServer(void)
{
	char *		serverId	= ilu_InventID();
	ilu_ObjectTable	objTab		= NULL;
	char *		transport	= getDefaultTransport();
	char *		protocol	= getDefaultProtocol(transport);
	ilu_Server	kserver;

/*
	fprintf(stderr, "creating default true server:");
	fprintf(stderr, " id='%s', transport='%s', protocol='%s'\n",
		serverId, transport, protocol);
*/
	if ((kserver = createTrueServer(serverId, objTab, transport,
	    protocol)) == 0)
		return 0;
	return ilusv_FromServer(kserver);
}

static PyObject *
getDefaultTrueServer(void)
{
	if (defaultTrueServer == 0)
		defaultTrueServer = createDefaultTrueServer();
	return defaultTrueServer;
}

static PyObject *
ilumod_DefaultServer(PyObject *self, PyObject *args)
{
	PyObject *	result;

	if (!PyArg_Parse(args, ""))
		return 0;
	if ((result = getDefaultTrueServer()) == 0)
		return 0;
	Py_INCREF(result);
	return result;
}

/********************************/

static IvObject *
getInstVars(PyInstanceObject *inst)
{
	PyObject *	iv;

	if ((iv = PyDict_GetItemString(inst->in_dict, nameVarInstVars)) == 0)
	{
		ilu_Class	kclass;

		if ((iv = iv_New()) == 0)
			return 0;
		if ((kclass = getKernelClass(inst->in_class)) == 0)
			return 0;
		((IvObject *) iv)->kclass = kclass;
		if (PyDict_SetItemString(inst->in_dict,
		    nameVarInstVars, iv) < 0)
		{
			Py_DECREF(iv);
			return 0;
		}
		Py_DECREF(iv); /* let dictionary have only reference */
	}
	else if (!iv_Check(iv))
	{
		char	msg[256];

		sprintf(msg, "instance's %s should be instvars",
			nameVarInstVars);
		PyErr_SetString(PyExc_TypeError, msg);
		return 0;
	}
	return (IvObject *) iv;
}

static char *
inventInstanceId(void)
{
	static int	nextId;
	char		buffer[32];

	sprintf(buffer, "%u", ++nextId);
	return stringDup(buffer);
}

static char *
getInstanceId(ilu_Class kclass, PyInstanceObject *inst)
{
	PyObject *ih;

	if (((ih = PyDict_GetItemString(inst->in_dict,
	    nameVarInstHandle)) == 0) || !PyString_Check(ih))
	  {
	    ih = PyString_FromString(inventInstanceId());
	    if (PyDict_SetItemString(inst->in_dict, nameVarInstHandle, ih) < 0)
	      {
		Py_DECREF(ih);
		return 0;
	      }
	  }
	return (PyString_AS_STRING((PyStringObject *) ih));
}

static void
cleanupUninterestingObjects (ilu_private junk)
{
  if (gcList != ILU_NIL && PyList_Check(gcList))
    {
      Py_DECREF(gcList);
      gcList = ILU_NIL;
    }
  gcAlarmSet = ilu_FALSE;
}

static void
setupGCCleanupAlarm ()
{
  if (gcAlarmSet)
    return;

  if (gcAlarm == ILU_NIL)
    gcAlarm = ilu_CreateAlarm();
  ilu_SetAlarm (gcAlarm,
		ilu_FineTime_Add (ilu_FineTime_Now(), plus1Minute),
		cleanupUninterestingObjects,
		ILU_NIL);
  gcAlarmSet = ilu_TRUE;
}

static void
removeObjFromGCList (PyObject *list, PyObject *inst)
{
  int len = PyList_Size(list);
  int i;

  for (i = 0;  i < len;  i++)
    if (PyList_GetItem(list, i) == inst)
      {
	PyList_SetSlice (list, i, i+1, Py_None);
	len = PyList_Size(list);
	i -= 1;
      }
}

static void
addObjToGCList (PyObject *list, PyObject *inst)
{
  int len = PyList_Size(list);
  int i;

  for (i = 0;  i < len;  i++)
    if (PyList_GetItem(list, i) == inst)
      return;	/* already there */
  PyList_Append (list, inst);
/*
  printf ("*** Added 0x%x to gc list\n", inst);
*/
}

static void
trackKernelInterest (ilu_Object kobj, int vi)
{
  /* called when ILU kernel gains or loses interest in a
     collectible object */
  PyInstanceObject *inst = ilu_GetLanguageSpecificObject(kobj);
  ilu_Class cl = ilu_ClassOfObject(kobj);

  if (inst != ILU_NIL && ((!ilu_TrueInstanceP(kobj)) || cl->cl_collectible))
    {
/*
      printf ("*** ILU kernel is %s in <%s> (0x%x, %s, %d)\n", vi ? "interested" : "not interested",
	      ilu_SBHOfObject(kobj), inst, ilu_TrueInstanceP(kobj) ? "T" : "S", Py_COUNT(inst));
*/
      if (vi)
	{
	  Py_INCREF(inst);
	  if (gcList != ILU_NIL && PyList_Check(gcList))
	    removeObjFromGCList (gcList, (PyObject *) inst);
	}
      else
	{
	  if (Py_COUNT(inst) < 2)
	    {
	      if (gcList == ILU_NIL)
		gcList = PyList_New(1);

	      addObjToGCList (gcList, (PyObject *) inst);
	      setupGCCleanupAlarm();
	    }
	  Py_DECREF(inst);
	}
    }
}

static ilu_Object
createKernelObject(IvObject *iv, PyInstanceObject *inst)
{
	char *		instId;
	char *		kih;
	ilu_Object	kobj;

	if ((instId = getInstanceId(iv->kclass, inst)) == 0)
		return 0;
	kih = stringDup(instId);
	kobj = ilu_FindOrCreateTrueObject(kih, iv->kserver, iv->kclass,
		(void *) inst);
	if (kobj == 0)
		PyErr_SetString(GeneralError, "FindOrCreateTrueObject failed");
	if (!(iv->kclass->cl_collectible))
	  Py_INCREF(inst);	/* hold onto Python true objects */
	return kobj;
}

static PyObject *
getInstanceSpecificServer(PyInstanceObject *inst)
{
	PyObject *	sv;

	if ((sv = PyObject_GetAttrString((PyObject *)inst, nameVarServer)) != 0)
	{
		if (ilusv_Check(sv))
			return sv;
	}
	else
		PyErr_Clear();
	return 0;
}

static IvObject *
getInstVarsWithKobj(PyInstanceObject *inst)
{
	IvObject *	iv;

	if ((iv = getInstVars(inst)) == 0)
		return 0;
	if (iv->kserver == 0)
	{
		PyObject *	sv;

		if ((sv = getInstanceSpecificServer(inst)) == 0 &&
		    (sv = getDefaultTrueServer()) == 0)
			return 0;
		iv->kserver = ((IlusvObject *) sv)->kserver;
	}
	ilu_EnterServer(iv->kserver, iv->kclass);
	if (iv->kobj == 0)
	{
		if ((iv->kobj = createKernelObject(iv, inst)) == 0)
		{
			ilu_ExitServer(iv->kserver, iv->kclass);
			return 0;
		}
	}
	return iv;
}

/********************************/

typedef struct
{
	PyObject *	objectOfIh;
	ilu_Server	kserver;
} OtRock;

static ilu_Object
getKobjFromOtObject(PyInstanceObject *inst, ilu_Server kserver)
{
	IvObject *	iv;

	if (!PyInstance_Check(inst))
	{
		fprintf(stderr, "object table: instance not returned\n");
		return 0;
	}

	if ((iv = getInstVars(inst)) == 0)
	{
		fprintf(stderr, "object table: returned instance is bad");
		return 0;
	}
	if (iv->kserver == 0)
		iv->kserver = kserver;
	else if (iv->kserver != kserver)
	{
		fprintf(stderr,
			"object table: returned instance has different server");
		return 0;
	}
	if (iv->kobj == 0)
		iv->kobj = createKernelObject(iv, inst);
	return iv->kobj;
}

static ilu_Object
callObjectOfIh(ilu_ObjectTable self, ilu_string ih)
{
	OtRock *	otr	= (OtRock *) self->ot_rock;
	PyObject *	args;
	PyObject *	inst;
	ilu_Object	kobj;

	if ((args = Py_BuildValue("(s)", ih)) == 0)
	{
		fprintf(stderr, "object table: couldn't build args\n");
		return 0;
	}
	inst = PyEval_CallObject(otr->objectOfIh, args);
	Py_DECREF(args);
	if (inst == 0)
	{
		handleCalloutException("object table: function");
		return 0;
	}
	kobj = getKobjFromOtObject((PyInstanceObject *) inst, otr->kserver);
	Py_DECREF(inst);
	return kobj;
}

static void
freeObjectTable(ilu_ObjectTable self)
{
	OtRock *	otr	= (OtRock *) self->ot_rock;

	Py_DECREF(otr->objectOfIh);
	PyMem_DEL(otr);
	PyMem_DEL(self);
}

static ilu_ObjectTable
createObjectTable(PyObject *objectOfIh)
{
	ilu_ObjectTable	objTab;
	OtRock *	otr;

	if ((objTab = PyMem_NEW(ilu_ObjectTable_s, 1)) == 0)
	{
		(void) PyErr_NoMemory();
		return 0;
	}
	if ((otr = PyMem_NEW(OtRock, 1)) == 0)
	{
		PyMem_DEL(objTab);
		(void) PyErr_NoMemory();
		return 0;
	}

	Py_INCREF(objectOfIh);
	otr->objectOfIh = objectOfIh;
	otr->kserver = 0;

	objTab->ot_object_of_ih = callObjectOfIh;
	objTab->ot_free_self = freeObjectTable;
	objTab->ot_rock = (ilu_private) otr;
	return objTab;
}

static void
setObjectTableServer(ilu_ObjectTable self, ilu_Server kserver)
{
	((OtRock *) self->ot_rock)->kserver = kserver;
}

/********************************/

static PyObject *
createSurrogateInstance(ilu_Object kobj)
{
	ilu_Class		kclass	= ilu_ClassOfObject(kobj);
	PyObject *		pycl;
	PyInstanceObject *	inst;
	IvObject *		iv;

	if ((pycl = findInClassRegistry(kclass)) == 0)
	{
		PyErr_SetString(GeneralError, "type not in registry");
		return 0;
	}
	if ((inst = (PyInstanceObject *) PyInstance_New(pycl, NULL)) == 0)
		return 0;
	if ((iv = getInstVars(inst)) == 0)
		return 0;
	iv->kserver = ilu_ServerOfObject(kobj);
	iv->kobj = kobj;
	return (PyObject *) inst;
}

static PyObject *
getPythonObject(ilu_Object kobj, ilu_Class kclass)
{
	ilu_Server	kserver	= ilu_ServerOfObject(kobj);
	PyObject *	pyobj;

	if ((pyobj = (PyObject *) ilu_GetLanguageSpecificObject(kobj)) == 0)
	{
		if ((pyobj = createSurrogateInstance(kobj)) == 0)
			return 0;
		ilu_RegisterLanguageSpecificObject(kobj, (void *) pyobj);
	}
	else
		Py_INCREF(pyobj);
	ilu_ExitServer(kserver, kclass);
	return pyobj;
}

static PyObject *
getPythonTrueObject(ilu_Object kobj, ilu_Class kclass)
{
	ilu_Server	kserver	= ilu_ServerOfObject(kobj);
	PyObject *	pyobj;

	if ((pyobj = (PyObject *) ilu_GetLanguageSpecificObject(kobj)) == 0)
		PyErr_SetString(GeneralError,
			"kernel object without true instance");
	else
		Py_INCREF(pyobj);
	ilu_ExitServer(kserver, kclass);
	return pyobj;
}

/********************************/

static int
ensureGcClient(void)
{
	static ilu_Server	gcClientKserver;
	static ilu_Object	gcClientKobj;
	char *			gcServerId;

	if (gcClientKserver == 0)
	{
		char *	trans	= getDefaultTransport();
		char *	proto	= getDefaultProtocol(trans);
		char *	sid	= ilu_InventID();

		if ((gcClientKserver = createTrueServer(sid, NULL,
		    trans, proto)) == 0)
		  return -1;
	}
	if (gcClientKobj == 0)
	{
		ilu_Class	kclass	= ilu_GetGcCallbackClass();
		char *		kih = ilu_InventID();

		ilu_EnterServer(gcClientKserver, kclass);

		/*
		** We're cheating here in giving something other than a
		** Python instance as the languageSpecificObject (4th arg).
		** But this should be okay as long as this gc object remains
		** inaccessible (its SBH is not public knowledge).
		*/
		gcClientKobj = ilu_FindOrCreateTrueObject(kih,
			gcClientKserver, kclass, (void *) &gcClientKobj);

		if (gcClientKobj != 0)
			ilu_SetGcClient(gcClientKobj);
		ilu_ExitServer(gcClientKserver, kclass);

		if (gcClientKobj == 0)
		{
			PyErr_SetString(GeneralError,
				"failed creating gc client object");
			return -1;
		}
	}
	return 0;
}

/********************************/

static PyObject *
ilumod_ObjectOfSBH(PyObject *self, PyObject *args)
{
	PyClassObject *	pycl;
	char *		sbh;
	char *		msTypeId;
	ilu_Class	kclass;
	ilu_Object	kobj;

	if (PyArg_Parse(args, "(Os)", &pycl, &sbh))
		msTypeId = 0;
	else
	{
		PyErr_Clear();
		if (!PyArg_Parse(args, "(Osz)", &pycl, &sbh, &msTypeId))
			return 0;
	}
	if (!PyClass_Check(pycl))
	{
		PyErr_SetString(PyExc_TypeError, "arg1 should be class");
		return 0;
	}
	if ((kclass = getKernelClass(pycl)) == 0)
		return 0;
	if (msTypeId == 0)
		msTypeId = kclass->cl_unique_id;
	if ((kobj = ilu_ObjectOfSBH(sbh, msTypeId, kclass)) == 0)
	{
		PyErr_SetString(GeneralError, "ilu_ObjectOfSBH failed");
		return 0;
	}
	return getPythonObject(kobj, kclass);
}

static PyObject *
ilumod_SBHOfObject(PyObject *self, PyObject *args)
{
	PyInstanceObject *	inst;
	char *			sbh;
	IvObject *		iv;

	if (!PyArg_Parse(args, "O", &inst))
		return 0;
	if (!PyInstance_Check(inst))
	{
		PyErr_SetString(PyExc_TypeError, "arg1 should be instance");
		return 0;
	}
	if ((iv = getInstVarsWithKobj(inst)) == 0)
		return 0;
	sbh = ilu_SBHOfObject(iv->kobj);
	ilu_ExitServer(iv->kserver, iv->kclass);
	if (sbh == 0)
	{
		PyErr_SetString(GeneralError, "ilu_SBHOfObject failed");
		return 0;
	}
	return PyString_FromString(sbh);
}

/********************************/

static PyObject *
ilumod_PublishObject (PyObject *self, PyObject *args)
{
  PyInstanceObject *inst;
  IvObject *iv;

  if (!PyArg_Parse(args, "O", &inst))
    return 0;
  if (!PyInstance_Check(inst))
    {
      PyErr_SetString(PyExc_TypeError, "arg1 should be instance");
      return 0;
    }
  if ((iv = getInstVarsWithKobj(inst)) == 0)
    return 0;
  iv->publish_proof = ilu_PublishObject(iv->kobj);
  if (iv->publish_proof != NULL)
    {
      Py_INCREF(Py_None);
      return Py_None;
    }
  else
    {
      PyErr_SetString(GeneralError, "ilu_PublishObject failed");
      return 0;
    }
}

static PyObject *
ilumod_WithdrawObject (PyObject *self, PyObject *args)
{
  PyInstanceObject *inst;
  IvObject *iv;

  if (!PyArg_Parse(args, "O", &inst))
    return 0;
  if (!PyInstance_Check(inst))
    {
      PyErr_SetString(PyExc_TypeError, "arg1 should be instance");
      return 0;
    }
  if ((iv = getInstVars(inst)) == 0)
    return 0;
  if (iv->kserver == 0 || iv->kclass == 0)
    {
      PyErr_SetString(GeneralError, "null class or kernel server for object");
      return 0;
    }
  if (iv->publish_proof == NULL)
    {
      PyErr_SetString(GeneralError, "not published");
      return 0;
    }
  ilu_EnterServer(iv->kserver, iv->kclass);
  if (ilu_WithdrawObject (iv->kobj, iv->publish_proof))
    {
      free(iv->publish_proof);
      iv->publish_proof = NULL;
      Py_INCREF(Py_None);
      return Py_None;
    }
  else
    {
      PyErr_SetString(GeneralError, "ilu_WithdrawObject failed");
      return 0;
    }
}

static PyObject *
ilumod_LookupObject (PyObject *self, PyObject *args)
{
  char *oid;
  PyClassObject *pycl;
  ilu_Class kclass;
  ilu_Object kobj;
  PyObject *pobj;

  if (!PyArg_Parse(args, "(sO)", &oid, &pycl))
    return 0;
  if (!PyClass_Check(pycl))
    {
      PyErr_SetString(PyExc_TypeError, "arg2 should be class");
      return 0;
    }
  if ((kclass = getKernelClass(pycl)) == 0)
    return 0;
  if ((kobj = ilu_LookupObject (oid, kclass)) == NULL)
    {
      Py_INCREF(Py_None);
      return Py_None;
    }
  else
    {
      pobj = getPythonObject (kobj, kclass);
      Py_INCREF(pobj);
      return pobj;
    }
}

static PyObject *
ilumod_ParseSBH(PyObject *self, PyObject *args)
{
	char *		sbh;
	char *		oid		= 0;
	char *		contactInfo	= 0;
	PyObject *	result;

	if (!PyArg_Parse(args, "s", &sbh))
		return 0;
	if (ilu_ParseStringBindingHandle(sbh, &oid, &contactInfo) == 0)
	{
		PyErr_SetString(GeneralError,
			"ilu_ParseStringBindingHandle failed");
		return 0;
	}
	result = Py_BuildValue("(ss)", oid, contactInfo);
	PyMem_XDEL(oid);
	PyMem_XDEL(contactInfo);
	return result;
}

static PyObject *
ilumod_ParseOID(PyObject *self, PyObject *args)
{
	char *		oid;
	char *		instanceHandle	= 0;
	char *		serverID	= 0;
	PyObject *	result;

	if (!PyArg_Parse(args, "s", &oid))
		return 0;
	if (ilu_ParseOID(oid, &instanceHandle, &serverID) == 0)
	{
		PyErr_SetString(GeneralError, "ilu_ParseOID failed");
		return 0;
	}
	result = Py_BuildValue("(ss)", instanceHandle, serverID);
	PyMem_XDEL(instanceHandle);
	PyMem_XDEL(serverID);
	return result;
}

static PyObject *
ilumod_PingObject (PyObject *self, PyObject *args)
{
  PyInstanceObject *inst;
  ilu_Error err;
  ilu_boolean ok;
  IvObject *iv;

  if (!PyArg_Parse(args, "O", &inst))
    return 0;
  if (!PyInstance_Check(inst))
    {
      PyErr_SetString(PyExc_TypeError, "arg1 should be instance");
      return 0;
    }
  if ((iv = getInstVarsWithKobj(inst)) == 0)
    return 0;

  /* now Inside(kobj->server, kobj->class) */

  err = ilu_DeltaHolds (iv->kobj, 1);
  ilu_ExitServer (iv->kserver, iv->kclass);
  ILU_ERR_SWITCH(err) {
    ILU_SUCCESS_CASE;
    ILU_ERR_CASE(GcRegFailed, v) {
      PyErr_SetString (GeneralError, "GC registration failed in ilu_DeltaHolds() call");
      return 0;
    }
    ILU_ERR_ELSE {
      PyErr_SetString (GeneralError, "Unknown error in ilu_DeltaHolds() call");
      return 0;
    }
  } ILU_ERR_ENDSWITCH;
  ok = ilu_PingObject(iv->kobj);
  ilu_EnterServer(iv->kserver, iv->kclass);
  err = ilu_DeltaHolds (iv->kobj, -1);
  ilu_ExitServer (iv->kserver, iv->kclass);
  ILU_ERR_SWITCH(err) {
    ILU_SUCCESS_CASE;
    ILU_ERR_CASE(GcRegFailed, v) {
      PyErr_SetString (GeneralError, "GC registration failed in ilu_DeltaHolds() call");
      return 0;
    }
    ILU_ERR_ELSE {
      PyErr_SetString (GeneralError, "Unknown error in ilu_DeltaHolds() call");
      return 0;
    }
  } ILU_ERR_ENDSWITCH;
  if (ok)
    return Py_True;
  else
    return Py_False;
}

/********************************/

static PyObject *
ilumod_RunMainLoop(PyObject *self, PyObject *args)
{
	if (!PyArg_Parse(args, ""))
		return 0;
	stopMainLoop = 0;
	ilu_RunMainLoop(&stopMainLoop);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_ExitMainLoop(PyObject *self, PyObject *args)
{
	if (!PyArg_Parse(args, ""))
		return 0;
	ilu_ExitMainLoop(&stopMainLoop);
	Py_INCREF(Py_None);
	return Py_None;
}

/********************************/

static PyObject *
callbackWithOneArg(PyObject *callback, PyObject *arg, char *what)
{
	PyObject *	argsTuple;
	PyObject *	result;

	if ((argsTuple = Py_BuildValue("(O)", arg)) == 0)
	{
		Py_DECREF(arg);
		return 0;
	}
	result = PyEval_CallObject(callback, argsTuple);
	Py_DECREF(argsTuple);
	if (result == 0)
	{
		char	culprit[100];

		sprintf(culprit, "%s callback", what);
		handleCalloutException(culprit);
	}
	return result;
}

static ilu_boolean
registerHandler(PyObject *callback, int fd, ilu_IOHandler handler,
	ilu_private rock, char *what)
{
	PyObject *	arg;
	PyObject *	result;
	ilu_boolean	stat;

	if ((arg = iohc_New(fd, handler, rock)) == 0 ||
	    (result = callbackWithOneArg(callback, arg, what)) == 0)
		return ilu_FALSE;
	stat = (result != Py_None);
	Py_DECREF(result);
	return stat;
}

static ilu_boolean
cancelHandler(PyObject *callback, int fd, char *what)
{
	PyObject *	arg;
	PyObject *	result;
	ilu_boolean	stat;

	if ((arg = PyInt_FromLong(fd)) == 0 ||
	    (result = callbackWithOneArg(callback, arg, what)) == 0)
		return ilu_FALSE;
	stat = (result != Py_None);
	Py_DECREF(result);
	return stat;
}

static PyObject *cbRegInp;
static PyObject *cbCanInp;
static PyObject *cbRegOut;
static PyObject *cbCanOut;
static PyObject *cbSetAlarm;
static PyObject *cbCanAlarm;

static ilu_boolean
registerInputHandler(int fd, ilu_IOHandler handler, ilu_private rock)
{
	return registerHandler(cbRegInp, fd, handler, rock, "reg input");
}

static ilu_boolean
cancelInputHandler(int fd)
{
	return cancelHandler(cbCanInp, fd, "cancel input");
}

static ilu_boolean
registerOutputHandler(int fd, ilu_IOHandler handler, ilu_private rock)
{
	return registerHandler(cbRegOut, fd, handler, rock, "reg output");
}

static ilu_boolean
cancelOutputHandler(int fd)
{
	return cancelHandler(cbCanOut, fd, "cancel output");
}

static void
setAlarmHandler(ilu_FineTime t, void (*proc)(ilu_FineTime t))
{
	PyObject *	arg;
	PyObject *	result;

	if ((arg = thc_New(t, proc)) == 0 ||
	    (result = callbackWithOneArg(cbSetAlarm, arg, "set alarm")) == 0)
		return;
	Py_DECREF(result);
}

static void
cancelAlarmHandler(void)
{
	PyObject *	argsTuple;
	PyObject *	result;

	if ((argsTuple = PyTuple_New(0)) == 0)
	{
		/* error */
		return;
	}
	result = PyEval_CallObject(cbCanAlarm, argsTuple);
	Py_DECREF(argsTuple);
	if (result == 0)
		handleCalloutException("cancel alarm callback");
	else
		Py_DECREF(result);
}

static PyObject *
ilumod_AddRegisterersToDefault(PyObject *self, PyObject *args)
{
	PyObject *	regInp;
	PyObject *	canInp;
	PyObject *	regOut;
	PyObject *	canOut;
	PyObject *	setAlarm;
	PyObject *	canAlarm;

	if (!PyArg_Parse(args, "(OOOOOO)", &regInp, &canInp, &regOut, &canOut,
	    &setAlarm, &canAlarm))
		return 0;
	if (!PyCallable_Check(regInp) || !PyCallable_Check(canInp) ||
	    !PyCallable_Check(regOut) || !PyCallable_Check(canOut) ||
	    !PyCallable_Check(setAlarm) || !PyCallable_Check(canAlarm))
	{
		PyErr_SetString(PyExc_TypeError,
			"all args should be callable");
		return 0;
	}
	cbRegInp = regInp;
	cbCanInp = canInp;
	cbRegOut = regOut;
	cbCanOut = canOut;
	cbSetAlarm = setAlarm;
	cbCanAlarm = canAlarm;
	if (ilu_AddRegisterersToDefault(registerInputHandler,
	    cancelInputHandler, registerOutputHandler, cancelOutputHandler,
	    setAlarmHandler, cancelAlarmHandler) != ilu_TRUE)
	{
		PyErr_SetString(GeneralError,
			"ilu_AddRegisterersToDefault failed");
		return 0;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

/********************************/

static PyObject *
longRealFromVector(
	int (*size)(PyObject *),
	PyObject *(*getItem)(PyObject *, int),
	PyObject *seq
	)
{
	ilu_longreal	lr;
	int		i;

	if ((*size)(seq) != ilulr_NBytesInLongReal)
	{
		PyErr_SetString(PyExc_TypeError, "arg1 has improper length");
		return 0;
	}
	for (i = 0; i < ilulr_NBytesInLongReal; i++)
	{
		PyObject *	item	= (*getItem)(seq, i);
		long		itemVal;

		if (!PyInt_Check(item))
		{
			PyErr_SetString(PyExc_TypeError,
				"item should be integer");
			return 0;
		}
		itemVal = PyInt_AsLong(item);
		if ((unsigned char) itemVal != itemVal)
		{
			PyErr_SetString(PyExc_ValueError,
				"item exceeds range of byte");
			return 0;
		}
		lr.b[i] = itemVal;
	}
	return ilulr_FromLongReal(lr);
}

static PyObject *
ilumod_LongReal(PyObject *self, PyObject *args)
{
	PyObject *	v;

	if (!PyArg_Parse(args, "O", &v))
		return 0;
	if (PyFloat_Check(v))
		return ilulr_FromDouble(PyFloat_AsDouble(v));
	else if (PyInt_Check(v))
		return ilulr_FromDouble((double) PyInt_AsLong(v));
	else if (PyList_Check(v))
		return longRealFromVector(PyList_Size, PyList_GetItem, v);
	else if (PyTuple_Check(v))
		return longRealFromVector(PyTuple_Size, PyTuple_GetItem, v);
	PyErr_SetString(PyExc_TypeError,
		"arg1 should be int, float, or sequence of bytes");
	return 0;
}

/********************************/

static PyObject *
ilumod_FineTime(PyObject *self, PyObject *args)
{
	PyObject *	value;
	ilu_FineTime	ft;

	if (!PyArg_Parse(args, "O", &value))
		return 0;
	if (PyFloat_Check(value))
		ft = ilu_FineTime_FromDouble(PyFloat_AsDouble(value));
	else if (PyInt_Check(value))
	{
		ft.ft_s = PyInt_AsLong(value);
		ft.ft_t = 0;
	}
	else
	{
		PyErr_SetString(PyExc_TypeError, "arg1 should be int or float");
		return 0;
	}
	return iluft_FromFineTime(ft);
}

static PyObject *
ilumod_FineTime_Now(PyObject *self, PyObject *args)
{
	ilu_FineTime	ft;

	if (!PyArg_Parse(args, ""))
		return 0;
	ft = ilu_FineTime_Now();
	return iluft_FromFineTime(ft);
}

static PyObject *
ilumod_CreateAlarm(PyObject *self, PyObject *args)
{
	if (!PyArg_Parse(args, ""))
		return 0;
	return ilual_New();
}

/********************************/

static char *
protoExceptionImage(ilu_ProtocolException pe)
{
	switch (pe)
	{
	case ilu_ProtocolException_Success:
		return "success";
	case ilu_ProtocolException_NoSuchClassAtServer:
		return "no such class at server";
	case ilu_ProtocolException_ClassVersionMismatch:
		return "class version mismatch";
	case ilu_ProtocolException_NoSuchMethodOnClass:
		return "no such method on class";
	case ilu_ProtocolException_GarbageArguments:
		return "garbage arguments";
	case ilu_ProtocolException_Unknown:
		return "unknown";
	case ilu_ProtocolException_LostConnection:
		return "lost connection";
	case ilu_ProtocolException_RequestRejected:
		return "request rejected";
	case ilu_ProtocolException_RequestTimeout:
		return "request timeout";
	default:
		return "?";
	}
}

static void
protoErr(ilu_ProtocolException pe)
{
	PyErr_SetString(ProtocolError, protoExceptionImage(pe));
}

/********************************/

static PyObject *
ilumod_BeginCall(PyObject *self, PyObject *args)
{
	PyInstanceObject *	inst;
	IvObject *		iv;
	ilu_Call		call;

	if (!PyArg_Parse(args, "O", &inst))
		return 0;
	if (!PyInstance_Check(inst))
	{
		PyErr_SetString(PyExc_TypeError, "arg1 should be instance");
		return 0;
	}
	if ((iv = getInstVars(inst)) == 0)
		return 0;
	if (iv->kserver == 0)
	{
		PyErr_SetString(GeneralError, "object has no server");
		return 0;
	}
	if ((call = ilu_BeginCall(iv->kserver)) == 0)
	{
		PyErr_SetString(GeneralError, "communication failure");
		return 0;
	}
	return iluca_FromCall(call);
}

static PyObject *
ilumod_FinishCall(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_FinishCall(ca->call);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_BeginRequest(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	IluclObject *	cl;
	int		methodIndex;
	long		argSize;
	ilu_Method	meth;

	if (!PyArg_Parse(args, "(OOil)", &ca, &cl, &methodIndex, &argSize))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!ilucl_Check(cl))
	{
		PyErr_SetString(PyExc_TypeError, "arg2 should be ilu_Class");
		return 0;
	}
	if (methodIndex < 1 || cl->c.cl_method_count < methodIndex)
	{
		PyErr_SetString(PyExc_TypeError,
			"arg3 (method index) out of bounds");
		return 0;
	}
	meth = cl->c.cl_methods + methodIndex - 1;
	if (ilu_BeginRequest(ca->call, &cl->c, meth, argSize) == ilu_FALSE)
	{
		protoErr(ilu_ProtocolException_Unknown);
		return 0;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_FinishRequest(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (ilu_FinishRequest(ca->call) == ilu_FALSE)
	{
		protoErr(ilu_ProtocolException_Unknown);
		return 0;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_GetReply(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	ilu_ProtocolException	pe;
	ilu_cardinal		scode;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	pe = ilu_GetReply(ca->call, &scode);
	if (pe != ilu_ProtocolException_Success)
	{
		protoErr(pe);
		return 0;
	}
	return PyInt_FromLong(scode);
}

static PyObject *
ilumod_ExceptionName(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	IluclObject *	cl;
	int		evalue;
	ilu_Method	meth;
	PyObject *	result;

	if (!PyArg_Parse(args, "(OOi)", &ca, &cl, &evalue))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!ilucl_Check(cl))
	{
		PyErr_SetString(PyExc_TypeError, "arg2 should be ilu_Class");
		return 0;
	}
	meth = ilu_MethodOfCall(ca->call);
	if (evalue < 0 || meth->me_exceptionCount < evalue)
	{
		PyErr_SetString(GeneralError, "unknown exception");
		return 0;
	}
	result = cl->excepts[meth - cl->c.cl_methods][evalue - 1];
	Py_INCREF(result);
	return result;
}

/********************************/

static PyObject *
ilumod_GetSingleton(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_Object	kobj;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	kobj = ilu_GetCallSingleton(ca->call);
	if (kobj == 0)
	{
		PyErr_SetString(GeneralError, "couldn't get singleton");
		return 0;
	}
	return getPythonTrueObject(kobj, ilu_ClassOfObject(kobj));
}

struct passport_list {
  ilu_Passport 		passport;
  PyObject *		python_version;
  struct passport_list *next;
};

static struct passport_list *all_passports = ILU_NIL;

/* before: conn not registered */
/* after:  conn not registered iff protocol concurrent */
static PyObject *
ilumod_RequestRead(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	ilu_Passport		pp;
	struct passport_list *	ppe;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}

	/* assumes single threading */
	ppe = (struct passport_list *) ilu_must_malloc(sizeof(struct passport_list));
	ppe->passport = ilu_CallerPassportOfCall(ca->call);
	ppe->python_version = Py_None;
	ppe->next = all_passports;
	all_passports = ppe;

	ilu_RequestRead(ca->call);
	enableRequestsOnConnIfConcurrent(ilu_ConnectionOfCall(ca->call));
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_CallerIdentity (PyObject *self, PyObject *args)
{
  if (all_passports == ILU_NIL)
    {
      PyErr_SetString (GeneralError, "no outstanding ILU calls");
      return 0;
    }

  if (all_passports->python_version == Py_None)
    {
      static ilu_AuthenticationScheme SunRPCUnixAS = ILU_NIL;
      PyObject *passport;

      if (SunRPCUnixAS == NULL)
	SunRPCUnixAS = ilu_FindAuthenticationSchemeByName (nameSunRPCAuth);
      if (SunRPCUnixAS == NULL)
	{
	  PyErr_SetString (GeneralError, "Can't find ILU SunRPC-UNIX authentication scheme");
	  return 0;
	}
      else
	{
	  PyObject *pp;
	  PyObject *auth;
	  PyObject *groups;
	  ilu_Principal p;
	  ilu_Error err = ILU_NIL;
	  int groupcount = 0;
	  int i;

	  p = ilu_GetPrincipal(all_passports->passport, SunRPCUnixAS, &err);
	  if (err == NULL)
	    {
	      ilu_AS_SunRPCUnix_UserID id = (ilu_AS_SunRPCUnix_UserID) p.ap_user_id;

	      /* create SunRPC identity record */

	      auth = PyDict_New ();
	      if (auth == ILU_NIL)
		{
		  PyErr_SetString (GeneralError, "Can't create passport object");
		  ilu_free(id);
		  return 0;
		}
	      PyDict_SetItemString (auth, nameSunRPCAuthUID, PyInt_FromLong((long) (id->au_UID)));
	      PyDict_SetItemString (auth, nameSunRPCAuthGID, PyInt_FromLong((long) (id->au_GID)));
	      PyDict_SetItemString (auth, nameSunRPCAuthHostname, PyString_FromString(id->au_hostname));
	      for (i = 0, groupcount = 0;  i < id->au_ngids;  i += 1)
		if (id->au_gids[i] != 0)
		  groupcount += 1;
	      if (groupcount == 0)
		groups = Py_None;
	      else
		{
		  groups = PyTuple_New(groupcount);
		  if (groups == 0)
		    {
		      PyErr_SetString (GeneralError, "Couldn't create tuple of user's groups");
		      free(id);
		      Py_DECREF(auth);
		      return 0;
		    }
		  for (i = 0, groupcount = 0;  i < id->au_ngids;  i++)
		    {
		      if (id->au_gids[i] != 0)
			{
			  PyTuple_SetItem (groups, groupcount, PyInt_FromLong((long) (id->au_gids[i])));
			  groupcount += 1;
			}
		    }
		}
	      PyDict_SetItemString (auth, nameSunRPCAuthGroups, groups);
	      free(id);

	      /* now create passport, and put Sun RPC identity into it */

	      pp = PyDict_New();
	      if (pp == ILU_NIL)
		{
		  PyErr_SetString (GeneralError, "Can't create passport object");
		  Py_DECREF(auth);
		  return (0);
		}
	      PyDict_SetItemString (pp, nameSunRPCAuth, auth);
	      Py_INCREF(pp);	/* for our list */
	      all_passports->python_version = pp;
	      Py_INCREF(pp);	/* for the caller */
	      return (pp);
	    }
	  else
	    {
	      PyErr_SetString (GeneralError, "couldn't get principal");
	      ilu_FreeError (err);
	      return (0);
	    }
	}
    }
  else
    {
      Py_INCREF(all_passports->python_version);
      return (all_passports->python_version);
    }
}

static void PopPassport (ilu_Call call)
{
  struct passport_list *e = all_passports;
  if (e != NULL)
    {
      all_passports = e->next;
      if (e->python_version != Py_None)
	Py_DECREF(e->python_version);
      e->python_version = 0;
      ilu_free(e);
    }
}

/* before: conn registered iff protocol concurrent */
/* after:  conn registered */
static PyObject *
ilumod_NoReply(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_Connection	conn;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	PopPassport (ca->call);
	conn = ilu_ConnectionOfCall(ca->call);
	ilu_NoReply(ca->call);
	enableRequestsOnConnIfNotConcurrent(conn);
	Py_INCREF(Py_None);
	return Py_None;
}

/* before: conn registered iff protocol concurrent
** after:  normal return   -> conn not registered,
**	   GeneralError    -> conn registered, or
**	   other exception -> conn registered iff protocol concurrent
*/
static PyObject *
ilumod_BeginReply(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	int		raises;
	long		argSize;
	ilu_Connection	conn;

	if (!PyArg_Parse(args, "(Oil)", &ca, &raises, &argSize))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	PopPassport (ca->call);
	conn = ilu_ConnectionOfCall(ca->call);
	disableRequestsOnConnIfConnConcurrent(conn);
	if (ilu_BeginReply(ca->call, raises, argSize) == ilu_FALSE)
	{
		enableRequestsOnConn(conn);
		PyErr_SetString(GeneralError, "reply error");
		return 0;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

/* before: conn not registered
** after:  normal return -> conn registered
**	   exception     -> conn not registered
*/
static PyObject *
ilumod_FinishReply(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_Connection	conn;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	conn = ilu_ConnectionOfCall(ca->call);
	ilu_FinishReply(ca->call);
	enableRequestsOnConn(conn);
	Py_INCREF(Py_None);
	return Py_None;
}

/* before: conn registered iff protocol concurrent
** after:  normal return   -> conn not registered
**	   GeneralError    -> conn registered
**	   other exception -> conn registered iff protocol concurrent
*/
static PyObject *
ilumod_BeginException(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	int		ecode;
	long		argSize;
	ilu_Connection	conn;

	if (!PyArg_Parse(args, "(Oil)", &ca, &ecode, &argSize))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	PopPassport (ca->call);
	conn = ilu_ConnectionOfCall(ca->call);
	disableRequestsOnConnIfConnConcurrent(conn);
	if (ilu_BeginException(ca->call, ecode, argSize) == ilu_FALSE)
	{
		enableRequestsOnConn(conn);
		PyErr_SetString(GeneralError, "exception error");
		return 0;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

/* before: conn not registered
** after:  normal return -> conn registered
**	   exception     -> conn not registered
*/
static PyObject *
ilumod_FinishException(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_Connection	conn;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	conn = ilu_ConnectionOfCall(ca->call);
	ilu_FinishException(ca->call);
	enableRequestsOnConn(conn);
	Py_INCREF(Py_None);
	return Py_None;
}

/* before: conn not registered iff protocol concurrent */
/* after:  conn registered (except for bad ca TypeError exception) */
static PyObject *
ilumod_CaughtUnexpectedException(PyObject *self, PyObject *args)
{
	PyObject *	excType	= PySys_GetObject("exc_type");
	PyObject *	excVal	= PySys_GetObject("exc_value");
	IlucaObject *	ca;
	ilu_Connection	conn;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	conn = ilu_ConnectionOfCall(ca->call);
	disableRequestsOnConnIfConnConcurrent(conn);
	if (excType != PyExc_SystemExit)
	{
		if (ilu_BeginException(ca->call, 0,
		    ilu_ProtocolException_Unknown) == ilu_FALSE)
		{
			enableRequestsOnConn(conn);
			PyErr_SetString(GeneralError,
				"ilu_BeginException failed");
			return 0;
		}
		ilu_FinishException(ca->call);
	}
	enableRequestsOnConn(conn);
	PyErr_SetObject(excType, excVal);
	return 0;
}


/************* begin marshalling/unmarshalling routines *************/

static PyObject *
ilumod_SizeOfObjectID(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	PyObject *		inst;
	int			isDiscrim;
	PyClassObject *		pycl;
	IvObject *		iv;
	ilu_Object		kobj;
	ilu_Class		kclass;
	long			size;

	if (!PyArg_Parse(args, "(OOiO)", &ca, &inst, &isDiscrim, &pycl))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!PyClass_Check(pycl))
	{
		PyErr_SetString(PyExc_TypeError, "arg4 should be class");
		return 0;
	}
	if ((kclass = getKernelClass(pycl)) == 0)
		return 0;
	if (!(PyInstance_Check(inst) ||
	    (inst == Py_None && !isDiscrim && kclass->cl_optional)))
	{
		PyErr_SetString(PyExc_TypeError, "arg2 should be instance");
		return 0;
	}
	if (inst != Py_None)
	{
		if ((iv = getInstVarsWithKobj((PyInstanceObject *) inst)) == 0)
			return 0;
		kobj = iv->kobj;
	}
	else
		kobj = 0;
	size = ilu_SizeOfObjectID(ca->call, kobj, isDiscrim, kclass);
	if ((PyObject *) inst != Py_None)
		ilu_ExitServer(iv->kserver, iv->kclass);
	return PyInt_FromLong(size);
}

static PyObject *
ilumod_OutputObjectID(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	PyObject *		inst;
	int			isDiscrim;
	PyClassObject *		pycl;
	IvObject *		iv;
	ilu_Object		kobj;
	ilu_Class		kclass;

	if (!PyArg_Parse(args, "(OOiO)", &ca, &inst, &isDiscrim, &pycl))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!PyClass_Check(pycl))
	{
		PyErr_SetString(PyExc_TypeError, "arg4 should be class");
		return 0;
	}
	if ((kclass = getKernelClass(pycl)) == 0)
		return 0;
	if (!(PyInstance_Check(inst) ||
	    (inst == Py_None && !isDiscrim && kclass->cl_optional)))
	{
		PyErr_SetString(PyExc_TypeError, "arg2 should be instance");
		return 0;
	}
	if (inst != Py_None)
	{
		if ((iv = getInstVarsWithKobj((PyInstanceObject *) inst)) == 0)
			return 0;
		kobj = iv->kobj;
	}
	else
		kobj = 0;
	ilu_OutputObjectID(ca->call, kobj, isDiscrim, kclass);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputObjectID(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	int		isDiscrim;
	PyClassObject *	pycl;
	ilu_Class	kclass;
	ilu_Object	kobj;

	if (!PyArg_Parse(args, "(OiO)", &ca, &isDiscrim, &pycl))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!PyClass_Check(pycl))
	{
		PyErr_SetString(PyExc_TypeError, "arg3 should be class");
		return 0;
	}
	if ((kclass = getKernelClass(pycl)) == 0)
		return 0;
	ilu_InputObjectID(ca->call, &kobj, isDiscrim, kclass);
	if (kobj == 0)
	{
		if (isDiscrim || !kclass->cl_optional)
		{
			PyErr_SetString(GeneralError,
				"ilu_InputObjectID failed");
			return 0;
		}
		Py_INCREF(Py_None);
		return Py_None;
	}
	if (isDiscrim)
		return getPythonTrueObject(kobj, kclass);
	return getPythonObject(kobj, kclass);
}

/********************************/

static PyObject *
ilumod_SizeOfShortInteger(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((ilu_shortinteger) value != value)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfShortInteger(ca->call, value));
}

static PyObject *
ilumod_OutputShortInteger(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((ilu_shortinteger) value != value)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}
	ilu_OutputShortInteger(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputShortInteger(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	ilu_shortinteger	value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputShortInteger(ca->call, &value);
	return PyInt_FromLong(value);
}

/********************************/

static PyObject *
ilumod_SizeOfInteger(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfInteger(ca->call, value));
}

static PyObject *
ilumod_OutputInteger(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_OutputInteger(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputInteger(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_integer	value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputInteger(ca->call, &value);
	return PyInt_FromLong(value);
}

/********************************/

/* Returns divmod(v, 2**32), after checking that v is in [lo, hi]. */
static PyObject *
decomposeLongInt(PyObject *v, char *hexLoBound, char *hexHiBound)
{
	int		isTooSmall;
	int		isTooLarge;
	PyObject *	liLo;
	PyObject *	liHi;
	PyObject *	li2_32;
	PyObject *	divmodTuple;

	if ((liLo = PYTHON_LONG_FROM_STRING(hexLoBound, 16)) == 0)
		return 0;
	isTooSmall = PyObject_Compare(v, liLo) < 0;
	Py_DECREF(liLo);
	if (isTooSmall)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}

	if ((liHi = PYTHON_LONG_FROM_STRING(hexHiBound, 16)) == 0)
		return 0;
	isTooLarge = PyObject_Compare(v, liHi) > 0;
	Py_DECREF(liHi);
	if (isTooLarge)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}

	if ((li2_32 = PYTHON_LONG_FROM_STRING("100000000", 16)) == 0)
		return 0;
	divmodTuple = (*v->ob_type->tp_as_number->nb_divmod)(v, li2_32);
	Py_DECREF(li2_32);
	return divmodTuple;
}

static int
getIluLongInteger(PyObject *v, ilu_longinteger *pValue)
{
	if (PyInt_Check(v))
	{
		long	value	= PyInt_AsLong(v);

		pValue->low = value & 0xffffffff;

		/* Behavior of a single 32-bit shift is not always defined. */
		pValue->high = value >> 16;
		pValue->high >>= 16;
		return 1;
	}
	if (PyLong_Check(v))
	{
		PyObject *	divmodTuple;

		if ((divmodTuple = decomposeLongInt(v,
		    "-8000000000000000", "7fffffffffffffff")) == 0)
			return 0;
		pValue->low = PyLong_AsDouble(PyTuple_GetItem(divmodTuple, 1));
		pValue->high = PyLong_AsDouble(PyTuple_GetItem(divmodTuple, 0));
		Py_DECREF(divmodTuple);
		return 1;
	}
	PyErr_SetString(PyExc_TypeError, strArg2ShouldBeIntOrLongInt);
	return 0;
}

static PyObject *
ilumod_SizeOfLongInteger(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	v;
	ilu_longinteger	value;

	if (!PyArg_Parse(args, "(OO)", &ca, &v))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!getIluLongInteger(v, &value))
		return 0;
	return PyInt_FromLong(ilu_SizeOfLongInteger(ca->call, value));
}

static PyObject *
ilumod_OutputLongInteger(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	v;
	ilu_longinteger	value;

	if (!PyArg_Parse(args, "(OO)", &ca, &v))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!getIluLongInteger(v, &value))
		return 0;
	ilu_OutputLongInteger(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputLongInteger(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_longinteger	value;
	int		sign;
	char		image[18];

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputLongInteger(ca->call, &value);
	if (value.high < 0)
	{
		sign = '-';
		value.low = 0xffffffff & - value.low;
		value.high = ~ value.high;
		if (value.low == 0)
			value.high += 1;
	}
	else
		sign = '+';
	sprintf(image, "%c%08lx%08lx", sign, (unsigned long) value.high,
	  (unsigned long) value.low);
	return PYTHON_LONG_FROM_STRING(image, 16);
}

/********************************/

static PyObject *
ilumod_SizeOfShortCardinal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((ilu_shortcardinal) value != value)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfShortCardinal(ca->call, value));
}

static PyObject *
ilumod_OutputShortCardinal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((ilu_shortcardinal) value != value)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}
	ilu_OutputShortCardinal(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputShortCardinal(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	ilu_shortcardinal	value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputShortCardinal(ca->call, &value);
	return PyInt_FromLong(value);
}

/********************************/

static int
getIluCardinal(PyObject *v, ilu_cardinal *pValue)
{
	if (PyInt_Check(v))
	{
		long	value	= PyInt_AsLong(v);

		if (value < 0)
		{
			PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
			return 0;
		}
		*pValue = value;
		return 1;
	}
	if (PyLong_Check(v))
	{
		double	value	= PyLong_AsDouble(v);

		if (value < 0 || value > (unsigned) 0xffffffff)
		{
			PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
			return 0;
		}
		*pValue = value;
		return 1;
	}
	PyErr_SetString(PyExc_TypeError, strArg2ShouldBeIntOrLongInt);
	return 0;
}

static PyObject *
ilumod_SizeOfCardinal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	v;
	ilu_cardinal	value;

	if (!PyArg_Parse(args, "(OO)", &ca, &v))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!getIluCardinal(v, &value))
		return 0;
	return PyInt_FromLong(ilu_SizeOfCardinal(ca->call, value));
}

static PyObject *
ilumod_OutputCardinal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	v;
	ilu_cardinal	value;

	if (!PyArg_Parse(args, "(OO)", &ca, &v))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!getIluCardinal(v, &value))
		return 0;
	ilu_OutputCardinal(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputCardinal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_cardinal	value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputCardinal(ca->call, &value);
	if ((long) value >= 0)
		return PyLong_FromLong(value); /* presumably faster */
	return PyLong_FromDouble((double) value);
}

/********************************/

static int
getIluLongCardinal(PyObject *v, ilu_longcardinal *pValue)
{
	if (PyInt_Check(v))
	{
		long	value	= PyInt_AsLong(v);

		if (value < 0)
		{
			PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
			return 0;
		}
		pValue->low = value;
		pValue->high = 0;
		return 1;
	}
	if (PyLong_Check(v))
	{
		PyObject *	divmodTuple;

		if ((divmodTuple = decomposeLongInt(v,
		    "0", "ffffffffffffffff")) == 0)
			return 0;
		pValue->low = PyLong_AsDouble(PyTuple_GetItem(divmodTuple, 1));
		pValue->high = PyLong_AsDouble(PyTuple_GetItem(divmodTuple, 0));
		Py_DECREF(divmodTuple);
		return 1;
	}
	PyErr_SetString(PyExc_TypeError, strArg2ShouldBeIntOrLongInt);
	return 0;
}

static PyObject *
ilumod_SizeOfLongCardinal(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	PyObject *		v;
	ilu_longcardinal	value;

	if (!PyArg_Parse(args, "(OO)", &ca, &v))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!getIluLongCardinal(v, &value))
		return 0;
	return PyInt_FromLong(ilu_SizeOfLongCardinal(ca->call, value));
}

static PyObject *
ilumod_OutputLongCardinal(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	PyObject *		v;
	ilu_longcardinal	value;

	if (!PyArg_Parse(args, "(OO)", &ca, &v))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!getIluLongCardinal(v, &value))
		return 0;
	ilu_OutputLongCardinal(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputLongCardinal(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	ilu_longcardinal	value;
	char			image[17];

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputLongCardinal(ca->call, &value);
	sprintf(image, "%08lx%08lx", (unsigned long) value.high,
	  (unsigned long) value.low);
	return PYTHON_LONG_FROM_STRING(image, 16);
}

/********************************/

static PyObject *
ilumod_SizeOfShortReal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	float		value;

	if (!PyArg_Parse(args, "(Of)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfShortReal(ca->call, value));
}

static PyObject *
ilumod_OutputShortReal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	float		value;

	if (!PyArg_Parse(args, "(Of)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_OutputShortReal(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputShortReal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	float		value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputShortReal(ca->call, &value);
	return PyFloat_FromDouble(value);
}

/********************************/

static PyObject *
ilumod_SizeOfReal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	double		value;

	if (!PyArg_Parse(args, "(Od)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfReal(ca->call, value));
}

static PyObject *
ilumod_OutputReal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	double		value;

	if (!PyArg_Parse(args, "(Od)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_OutputReal(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputReal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	double		value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputReal(ca->call, &value);
	return PyFloat_FromDouble(value);
}

/********************************/

static PyObject *
ilumod_SizeOfLongReal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	IlulrObject *	v;

	if (!PyArg_Parse(args, "(OO)", &ca, &v))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!ilulr_Check(v))
	{
		PyErr_SetString(PyExc_TypeError, "arg2 should be ilu_longreal");
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfLongReal(ca->call,
					ilulr_AS_LONGREAL(v)));
}

static PyObject *
ilumod_OutputLongReal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	IlulrObject *	v;

	if (!PyArg_Parse(args, "(OO)", &ca, &v))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!ilulr_Check(v))
	{
		PyErr_SetString(PyExc_TypeError, "arg2 should be ilu_longreal");
		return 0;
	}
	ilu_OutputLongReal(ca->call, ilulr_AS_LONGREAL(v));
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputLongReal(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_longreal	value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputLongReal(ca->call, &value);
	return ilulr_FromLongReal(value);
}

/********************************/

static PyObject *
ilumod_SizeOfEnum(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	int		value;

	if (!PyArg_Parse(args, "(Oi)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfEnum(ca->call, value));
}

static PyObject *
ilumod_OutputEnum(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	int		value;

	if (!PyArg_Parse(args, "(Oi)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_OutputEnum(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputEnum(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	ilu_shortcardinal	value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputEnum(ca->call, &value);
	return PyInt_FromLong(value);
}

/********************************/

static PyObject *
ilumod_SizeOfCharacter(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((ilu_character) value != value)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfCharacter(ca->call, value));
}

static PyObject *
ilumod_OutputCharacter(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((ilu_character) value != value)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}
	ilu_OutputCharacter(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputCharacter(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_character	value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputCharacter(ca->call, &value);
	return PyInt_FromLong(value);
}

/********************************/

static PyObject *
ilumod_SizeOfByte(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((ilu_byte) value != value)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfByte(ca->call, value));
}

static PyObject *
ilumod_OutputByte(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((ilu_byte) value != value)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}
	ilu_OutputByte(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputByte(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_byte	value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputByte(ca->call, &value);
	return PyInt_FromLong(value);
}

/********************************/

static PyObject *
ilumod_SizeOfBoolean(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (value != 0 && value != 1)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfBoolean(ca->call, value));
}

static PyObject *
ilumod_OutputBoolean(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		value;

	if (!PyArg_Parse(args, "(Ol)", &ca, &value))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (value != 0 && value != 1)
	{
		PyErr_SetString(PyExc_ValueError, strArg2OutOfRange);
		return 0;
	}
	ilu_OutputBoolean(ca->call, value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputBoolean(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_boolean	value;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputBoolean(ca->call, &value);
	return PyInt_FromLong(value);
}

/********************************/

static ilu_bytes
vectorToBytes(PyObject *vec, long *pLen, int *pAlloced)
{
	PyObject *	(*getItem)(PyObject *, int);
	int		len;
	ilu_bytes	value;
	int		i;

	if (PyString_Check(vec))
	{
		*pLen = PyString_Size(vec);
		*pAlloced = 0;
		return (ilu_bytes) PyString_AsString(vec);
	}
	else if (PyList_Check(vec))
	{
		getItem = PyList_GetItem;
		len = PyList_Size(vec);
	}
	else if (PyTuple_Check(vec))
	{
		getItem = PyTuple_GetItem;
		len = PyTuple_Size(vec);
	}
	else
	{
		PyErr_SetString(PyExc_TypeError,
			"arg2 should be string, list, or tuple");
		return 0;
	}
	if ((value = PyMem_NEW(ilu_byte, len)) == 0)
	{
		(void) PyErr_NoMemory();
		return 0;
	}
	for (i = 0; i < len; i++)
	{
		PyObject *	item	= (*getItem)(vec, i);
		long		itemVal;

		if (!PyInt_Check(item))
		{
			PyErr_SetString(PyExc_TypeError,
				"item should be integer");
			PyMem_DEL(value);
			return 0;
		}
		itemVal = PyInt_AsLong(item);
		if ((ilu_byte) itemVal != itemVal)
		{
			PyErr_SetString(PyExc_ValueError,
				"item exceeds range of byte");
			PyMem_DEL(value);
			return 0;
		}
		value[i] = itemVal;
	}
	*pLen = len;
	*pAlloced = 1;
	return value;
}

static PyObject *
listFromBytes(ilu_bytes value, int len)
{
	PyObject *	list;
	int		i;

	if ((list = PyList_New(len)) == 0)
		return 0;
	for (i = 0; i < len; i++)
	{
		PyObject *	item;
		
		if ((item = PyInt_FromLong(value[i])) == 0)
		{
			Py_DECREF(list);
			return 0;
		}
		if (PyList_SetItem(list, i, item) < 0)
		{
			Py_DECREF(list);
			return 0;
		}
	}
	return list;
}

static PyObject *
ilumod_SizeOfBytes(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	vec;
	long		limit;
	long		len;
	int		alloced;
	ilu_bytes	value;
	long		size;

	if (!PyArg_Parse(args, "(OOl)", &ca, &vec, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((value = vectorToBytes(vec, &len, &alloced)) == 0)
		return 0;
	if (limit > 0 && len > limit)
	{
		PyErr_SetString(PyExc_TypeError, strListTooLong);
		if (alloced)
			PyMem_DEL(value);
		return 0;
	}
	size = ilu_SizeOfBytes(ca->call, value, len, limit, ilu_FALSE);
	if (alloced)
		PyMem_DEL(value);
	return PyInt_FromLong(size);
}

static PyObject *
ilumod_OutputBytes(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	vec;
	long		limit;
	long		len;
	int		alloced;
	ilu_bytes	value;

	if (!PyArg_Parse(args, "(OOl)", &ca, &vec, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((value = vectorToBytes(vec, &len, &alloced)) == 0)
		return 0;
	if (limit > 0 && len > limit)
	{
		PyErr_SetString(PyExc_TypeError, strListTooLong);
		if (alloced)
			PyMem_DEL(value);
		return 0;
	}
	ilu_OutputBytes(ca->call, value, len, limit, ilu_FALSE);
	if (alloced)
		PyMem_DEL(value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputBytes(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		limit;
	ilu_bytes	value;
	ilu_cardinal	len;
	PyObject *	list;

	if (!PyArg_Parse(args, "(Ol)", &ca, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	value = 0;
	ilu_InputBytes(ca->call, &value, &len, limit, ilu_FALSE);
	list = listFromBytes(value, len);
	PyMem_XDEL(value);
	return list;
}

/********************************/

static PyObject *
ilumod_SizeOfString(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	str;
	long		limit;
	ilu_string	value;
	ilu_cardinal	len;

	if (!PyArg_Parse(args, "(OOl)", &ca, &str, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!PyString_Check(str))
	{
		PyErr_SetString(PyExc_TypeError, "arg2 should be string");
		return 0;
	}
	value = PyString_AsString(str);
	len = PyString_Size(str);
	return PyInt_FromLong(ilu_SizeOfString(ca->call, value, len, limit,
		ilu_FALSE));
}

static PyObject *
ilumod_OutputString(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	str;
	long		limit;
	ilu_string	value;
	ilu_cardinal	len;

	if (!PyArg_Parse(args, "(OOl)", &ca, &str, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if (!PyString_Check(str))
	{
		PyErr_SetString(PyExc_TypeError, "arg2 should be string");
		return 0;
	}
	value = PyString_AsString(str);
	len = PyString_Size(str);
	ilu_OutputString(ca->call, value, len, limit, ilu_FALSE);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputString(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		limit;
	ilu_string	value;
	ilu_cardinal	len;
	PyObject *	str;

	if (!PyArg_Parse(args, "(Ol)", &ca, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	value = 0;
	ilu_InputString(ca->call, &value, &len, limit, ilu_FALSE);
	str = PyString_FromStringAndSize(value, len);
	if (value)
		free(value);
	return str;
}

/********************************/

static ilu_wstring
vectorToWString(PyObject *vec, long *pLen)
{
	PyObject *	(*getItem)(PyObject *, int);
	int		len;
	ilu_wstring	value;
	int		i;

	if (PyString_Check(vec))
	{
		char *	string	= PyString_AsString(vec);

		len = PyString_Size(vec);
		if ((value = PyMem_NEW(ilu_character, len)) == 0)
		{
			(void) PyErr_NoMemory();
			return 0;
		}
		for (i = 0; i < len; i++)
			value[i] = string[i];
		*pLen = len;
		return value;
	}
	else if (PyList_Check(vec))
	{
		getItem = PyList_GetItem;
		len = PyList_Size(vec);
	}
	else if (PyTuple_Check(vec))
	{
		getItem = PyTuple_GetItem;
		len = PyTuple_Size(vec);
	}
	else
	{
		PyErr_SetString(PyExc_TypeError,
			"arg2 should be string, list, or tuple");
		return 0;
	}
	if ((value = PyMem_NEW(ilu_character, len)) == 0)
	{
		(void) PyErr_NoMemory();
		return 0;
	}
	for (i = 0; i < len; i++)
	{
		PyObject *	item	= (*getItem)(vec, i);
		long		itemVal;

		if (!PyInt_Check(item))
		{
			PyErr_SetString(PyExc_TypeError,
				"item should be integer");
			PyMem_DEL(value);
			return 0;
		}
		itemVal = PyInt_AsLong(item);
		if ((ilu_character) itemVal != itemVal)
		{
			PyErr_SetString(PyExc_ValueError,
				"item exceeds range of ilu character");
			PyMem_DEL(value);
			return 0;
		}
		value[i] = itemVal;
	}
	*pLen = len;
	return value;
}

static PyObject *
listFromWString(ilu_wstring value, int len)
{
	PyObject *	list;
	int		i;

	if ((list = PyList_New(len)) == 0)
		return 0;
	for (i = 0; i < len; i++)
	{
		PyObject *	item;
		
		if ((item = PyInt_FromLong(value[i])) == 0)
		{
			Py_DECREF(list);
			return 0;
		}
		if (PyList_SetItem(list, i, item) < 0)
		{
			Py_DECREF(list);
			return 0;
		}
	}
	return list;
}

static PyObject *
ilumod_SizeOfWString(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	vec;
	long		limit;
	long		len;
	ilu_wstring	value;
	long		size;

	if (!PyArg_Parse(args, "(OOl)", &ca, &vec, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((value = vectorToWString(vec, &len)) == 0)
		return 0;
	if (limit > 0 && len > limit)
	{
		PyErr_SetString(PyExc_TypeError, strListTooLong);
		PyMem_DEL(value);
		return 0;
	}
	size = ilu_SizeOfWString(ca->call, value, len, limit,
		NULL, NULL, ilu_FALSE);
	PyMem_DEL(value);
	return PyInt_FromLong(size);
}

static PyObject *
ilumod_OutputWString(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	vec;
	long		limit;
	long		len;
	ilu_wstring	value;

	if (!PyArg_Parse(args, "(OOl)", &ca, &vec, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((value = vectorToWString(vec, &len)) == 0)
		return 0;
	if (limit > 0 && len > limit)
	{
		PyErr_SetString(PyExc_TypeError, strListTooLong);
		PyMem_DEL(value);
		return 0;
	}
	ilu_OutputWString(ca->call, value, len, limit,
		NULL, NULL, ilu_FALSE);
	PyMem_DEL(value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputWString(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		limit;
	ilu_wstring	value;
	ilu_cardinal	len;
	PyObject *	list;

	if (!PyArg_Parse(args, "(Ol)", &ca, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	value = 0;
	ilu_InputWString(ca->call, &value, &len, limit, ilu_FALSE);
	list = listFromWString(value, len);
	PyMem_XDEL(value);
	return list;
}

/********************************/

static PyObject *
ilumod_SizeOfOpaque(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	vec;
	long		bound;
	long		len;
	int		alloced;
	ilu_bytes	value;
	long		size;

	if (!PyArg_Parse(args, "(OOl)", &ca, &vec, &bound))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((value = vectorToBytes(vec, &len, &alloced)) == 0)
		return 0;
	if (len != bound)
	{
		PyErr_SetString(PyExc_TypeError, strListDoesntMatch);
		if (alloced)
			PyMem_DEL(value);
		return 0;
	}
	size = ilu_SizeOfOpaque(ca->call, value, bound, ilu_FALSE);
	if (alloced)
		PyMem_DEL(value);
	return PyInt_FromLong(size);
}

static PyObject *
ilumod_OutputOpaque(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	vec;
	long		bound;
	long		len;
	int		alloced;
	ilu_bytes	value;

	if (!PyArg_Parse(args, "(OOl)", &ca, &vec, &bound))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((value = vectorToBytes(vec, &len, &alloced)) == 0)
		return 0;
	if (len != bound)
	{
		PyErr_SetString(PyExc_TypeError, strListDoesntMatch);
		if (alloced)
			PyMem_DEL(value);
		return 0;
	}
	ilu_OutputOpaque(ca->call, value, bound, ilu_FALSE);
	if (alloced)
		PyMem_DEL(value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputOpaque(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		bound;
	ilu_bytes	value;
	PyObject *	list;

	if (!PyArg_Parse(args, "(Ol)", &ca, &bound))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	value = 0;
	ilu_InputOpaque(ca->call, &value, bound, ilu_FALSE);
	list = listFromBytes(value, bound);
	PyMem_XDEL(value);
	return list;
}

/********************************/

static PyObject *
ilumod_SizeOfStringVec(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	vec;
	long		bound;
	long		len;
	int		alloced;
	ilu_string	value;
	long		size;

	if (!PyArg_Parse(args, "(OOl)", &ca, &vec, &bound))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((value = (ilu_string) vectorToBytes(vec, &len, &alloced)) == 0)
		return 0;
	if (len != bound)
	{
		PyErr_SetString(PyExc_TypeError, strListDoesntMatch);
		if (alloced)
			PyMem_DEL(value);
		return 0;
	}
	size = ilu_SizeOfStringVec(ca->call, value, bound, ilu_FALSE);
	if (alloced)
		PyMem_DEL(value);
	return PyInt_FromLong(size);
}

static PyObject *
ilumod_OutputStringVec(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	vec;
	long		bound;
	long		len;
	int		alloced;
	ilu_string	value;

	if (!PyArg_Parse(args, "(OOl)", &ca, &vec, &bound))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((value = (ilu_string) vectorToBytes(vec, &len, &alloced)) == 0)
		return 0;
	if (len != bound)
	{
		PyErr_SetString(PyExc_TypeError, strListDoesntMatch);
		if (alloced)
			PyMem_DEL(value);
		return 0;
	}
	ilu_OutputStringVec(ca->call, value, bound, ilu_FALSE);
	if (alloced)
		PyMem_DEL(value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputStringVec(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		bound;
	ilu_string	value;
	PyObject *	list;

	if (!PyArg_Parse(args, "(Ol)", &ca, &bound))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	value = 0;
	ilu_InputStringVec(ca->call, &value, bound, ilu_FALSE);
	list = listFromBytes((ilu_bytes) value, bound);
	PyMem_XDEL(value);
	return list;
}

/********************************/

static PyObject *
ilumod_SizeOfWStringVec(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	vec;
	long		bound;
	long		len;
	ilu_wstring	value;
	long		size;

	if (!PyArg_Parse(args, "(OOl)", &ca, &vec, &bound))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((value = vectorToWString(vec, &len)) == 0)
		return 0;
	if (len != bound)
	{
		PyErr_SetString(PyExc_TypeError, strListDoesntMatch);
		PyMem_DEL(value);
		return 0;
	}
	size = ilu_SizeOfWStringVec(ca->call, value, bound,
		NULL, NULL, ilu_FALSE);
	PyMem_DEL(value);
	return PyInt_FromLong(size);
}

static PyObject *
ilumod_OutputWStringVec(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	PyObject *	vec;
	long		bound;
	long		len;
	ilu_wstring	value;

	if (!PyArg_Parse(args, "(OOl)", &ca, &vec, &bound))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	if ((value = vectorToWString(vec, &len)) == 0)
		return 0;
	if (len != bound)
	{
		PyErr_SetString(PyExc_TypeError, strListDoesntMatch);
		PyMem_DEL(value);
		return 0;
	}
	ilu_OutputWStringVec(ca->call, value, bound,
		NULL, NULL, ilu_FALSE);
	PyMem_DEL(value);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputWStringVec(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		bound;
	ilu_wstring	value;
	PyObject *	list;

	if (!PyArg_Parse(args, "(Ol)", &ca, &bound))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	value = 0;
	ilu_InputWStringVec(ca->call, &value, bound, ilu_FALSE);
	list = listFromWString(value, bound);
	PyMem_XDEL(value);
	return list;
}

/********************************/

static PyObject *
ilumod_SizeOfOptional(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	int		present;

	if (!PyArg_Parse(args, "(Oi)", &ca, &present))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfOptional(ca->call, present));
}

static PyObject *
ilumod_OutputOptional(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	int		present;

	if (!PyArg_Parse(args, "(Oi)", &ca, &present))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_OutputOptional(ca->call, present);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputOptional(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_boolean	present;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputOptional(ca->call, &present);
	if (present)
		return PyInt_FromLong(present);
	Py_INCREF(Py_None);
	return Py_None;
}

/********************************/

static PyObject *
ilumod_SizeOfUnion(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		discrim;

	if (!PyArg_Parse(args, "(Ol)", &ca, &discrim))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	return PyInt_FromLong(
		ilu_SizeOfUnion(ca->call, discrim, ilu_FALSE, ilu_TRUE));
}

static PyObject *
ilumod_OutputUnion(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		discrim;

	if (!PyArg_Parse(args, "(Ol)", &ca, &discrim))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_OutputUnion(ca->call, discrim, ilu_FALSE, ilu_TRUE);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputUnion(PyObject *self, PyObject *args)
{
	IlucaObject *		ca;
	ilu_shortcardinal	discrim;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputUnion(ca->call, &discrim, ilu_FALSE, NULL);
	return PyInt_FromLong(discrim);
}

static PyObject *
ilumod_EndUnion(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_EndUnion(ca->call);
	Py_INCREF(Py_None);
	return Py_None;
}

/********************************/

static PyObject *
ilumod_SizeOfArray(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		length;

	if (!PyArg_Parse(args, "(Ol)", &ca, &length))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	return PyInt_FromLong(
		ilu_SizeOfArray(ca->call, length, ilu_FALSE, ilu_TRUE));
}

static PyObject *
ilumod_OutputArray(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		length;

	if (!PyArg_Parse(args, "(Ol)", &ca, &length))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_OutputArray(ca->call, length, ilu_FALSE, ilu_TRUE);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputArray(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputArray(ca->call, ilu_FALSE, NULL);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_EndArray(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_EndArray(ca->call);
	Py_INCREF(Py_None);
	return Py_None;
}

/********************************/

static PyObject *
ilumod_SizeOfRecord(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	return PyInt_FromLong(ilu_SizeOfRecord(ca->call, ilu_FALSE, ilu_TRUE));
}

static PyObject *
ilumod_OutputRecord(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_OutputRecord(ca->call, ilu_FALSE, ilu_TRUE);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputRecord(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputRecord(ca->call, ilu_FALSE, NULL);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_EndRecord(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_EndRecord(ca->call);
	Py_INCREF(Py_None);
	return Py_None;
}

/********************************/

static PyObject *
ilumod_SizeOfSequence(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		len;
	long		limit;

	if (!PyArg_Parse(args, "(Oll)", &ca, &len, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	return PyInt_FromLong(
		ilu_SizeOfSequence(ca->call, len, limit, ilu_FALSE, ilu_TRUE));
}

static PyObject *
ilumod_OutputSequence(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	long		len;
	long		limit;

	if (!PyArg_Parse(args, "(Oll)", &ca, &len, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_OutputSequence(ca->call, len, limit, ilu_FALSE, ilu_TRUE);
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
ilumod_InputSequence(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;
	ilu_cardinal	len;
	long		limit;

	if (!PyArg_Parse(args, "(Ol)", &ca, &limit))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_InputSequence(ca->call, &len, limit, ilu_FALSE, NULL);
	return PyInt_FromLong((long) len);
}

static PyObject *
ilumod_EndSequence(PyObject *self, PyObject *args)
{
	IlucaObject *	ca;

	if (!PyArg_Parse(args, "O", &ca))
		return 0;
	if (!iluca_Check(ca))
	{
		PyErr_SetString(PyExc_TypeError, strArg1ShouldBeIluCall);
		return 0;
	}
	ilu_EndSequence(ca->call);
	Py_INCREF(Py_None);
	return Py_None;
}

/************* end marshalling/unmarshalling routines *************/


static PyMethodDef ilumod_methods[] =
{
    {	"SetDebugLevel",		ilumod_SetDebugLevel		},
    {	"SetDebugLevelViaString",	ilumod_SetDebugLevelViaString	},

    {	"FormClassRecord",		ilumod_FormClassRecord		},
    {	"RegisterClass",		ilumod_RegisterClass		},
    {	"RegisterSkeletons",		ilumod_RegisterSkeletons	},
    {	"CreateServer",			ilumod_CreateServer		},
    {	"DefaultServer",		ilumod_DefaultServer		},
    {	"ObjectOfSBH",			ilumod_ObjectOfSBH		},
    {	"SBHOfObject",			ilumod_SBHOfObject		},

    {	"PublishObject",		ilumod_PublishObject		},
    {	"WithdrawObject",		ilumod_WithdrawObject		},
    {	"LookupObject",			ilumod_LookupObject		},
    {   "PingObject",			ilumod_PingObject		},
    {	"ParseSBH",			ilumod_ParseSBH			},
    {	"ParseOID",			ilumod_ParseOID			},

    {	"RunMainLoop",			ilumod_RunMainLoop		},
    {	"ExitMainLoop",			ilumod_ExitMainLoop		},

    {	"AddRegisterersToDefault",	ilumod_AddRegisterersToDefault	},

    {	"LongReal",			ilumod_LongReal			},

    {	"FineTime",			ilumod_FineTime			},
    {	"FineTime_Now",			ilumod_FineTime_Now		},
    {	"CreateAlarm",			ilumod_CreateAlarm		},

    {	"BeginCall",			ilumod_BeginCall		},
    {	"FinishCall",			ilumod_FinishCall		},
    {	"BeginRequest",			ilumod_BeginRequest		},
    {	"FinishRequest",		ilumod_FinishRequest		},
    {	"GetReply",			ilumod_GetReply			},
    {	"ExceptionName",		ilumod_ExceptionName		},

    {	"GetSingleton",			ilumod_GetSingleton		},
    {	"RequestRead",			ilumod_RequestRead		},
    {	"CallerIdentity",		ilumod_CallerIdentity		},
    {	"NoReply",			ilumod_NoReply			},
    {	"BeginReply",			ilumod_BeginReply		},
    {	"FinishReply",			ilumod_FinishReply		},
    {	"BeginException",		ilumod_BeginException		},
    {	"FinishException",		ilumod_FinishException		},
    {	"CaughtUnexpectedException",	ilumod_CaughtUnexpectedException},

    {	"SizeOfObjectID",		ilumod_SizeOfObjectID		},
    {	"OutputObjectID",		ilumod_OutputObjectID		},
    {	"InputObjectID",		ilumod_InputObjectID		},

    {	"SizeOfShortInteger",		ilumod_SizeOfShortInteger	},
    {	"OutputShortInteger",		ilumod_OutputShortInteger	},
    {	"InputShortInteger",		ilumod_InputShortInteger	},

    {	"SizeOfInteger",		ilumod_SizeOfInteger		},
    {	"OutputInteger",		ilumod_OutputInteger		},
    {	"InputInteger",			ilumod_InputInteger		},

    {	"SizeOfLongInteger",		ilumod_SizeOfLongInteger	},
    {	"OutputLongInteger",		ilumod_OutputLongInteger	},
    {	"InputLongInteger",		ilumod_InputLongInteger		},

    {	"SizeOfShortCardinal",		ilumod_SizeOfShortCardinal	},
    {	"OutputShortCardinal",		ilumod_OutputShortCardinal	},
    {	"InputShortCardinal",		ilumod_InputShortCardinal	},

    {	"SizeOfCardinal",		ilumod_SizeOfCardinal		},
    {	"OutputCardinal",		ilumod_OutputCardinal		},
    {	"InputCardinal",		ilumod_InputCardinal		},

    {	"SizeOfLongCardinal",		ilumod_SizeOfLongCardinal	},
    {	"OutputLongCardinal",		ilumod_OutputLongCardinal	},
    {	"InputLongCardinal",		ilumod_InputLongCardinal	},

    {	"SizeOfShortReal",		ilumod_SizeOfShortReal		},
    {	"OutputShortReal",		ilumod_OutputShortReal		},
    {	"InputShortReal",		ilumod_InputShortReal		},

    {	"SizeOfReal",			ilumod_SizeOfReal		},
    {	"OutputReal",			ilumod_OutputReal		},
    {	"InputReal",			ilumod_InputReal		},

    {	"SizeOfLongReal",		ilumod_SizeOfLongReal		},
    {	"OutputLongReal",		ilumod_OutputLongReal		},
    {	"InputLongReal",		ilumod_InputLongReal		},

    {	"SizeOfEnum",			ilumod_SizeOfEnum		},
    {	"OutputEnum",			ilumod_OutputEnum		},
    {	"InputEnum",			ilumod_InputEnum		},

    {	"SizeOfCharacter",		ilumod_SizeOfCharacter		},
    {	"OutputCharacter",		ilumod_OutputCharacter		},
    {	"InputCharacter",		ilumod_InputCharacter		},

    {	"SizeOfByte",			ilumod_SizeOfByte		},
    {	"OutputByte",			ilumod_OutputByte		},
    {	"InputByte",			ilumod_InputByte		},

    {	"SizeOfBoolean",		ilumod_SizeOfBoolean		},
    {	"OutputBoolean",		ilumod_OutputBoolean		},
    {	"InputBoolean",			ilumod_InputBoolean		},

    {	"SizeOfBytes",			ilumod_SizeOfBytes		},
    {	"OutputBytes",			ilumod_OutputBytes		},
    {	"InputBytes",			ilumod_InputBytes		},

    {	"SizeOfString",			ilumod_SizeOfString		},
    {	"OutputString",			ilumod_OutputString		},
    {	"InputString",			ilumod_InputString		},

    {	"SizeOfWString",		ilumod_SizeOfWString		},
    {	"OutputWString",		ilumod_OutputWString		},
    {	"InputWString",			ilumod_InputWString		},

    {	"SizeOfOpaque",			ilumod_SizeOfOpaque		},
    {	"OutputOpaque",			ilumod_OutputOpaque		},
    {	"InputOpaque",			ilumod_InputOpaque		},

    {	"SizeOfStringVec",		ilumod_SizeOfStringVec		},
    {	"OutputStringVec",		ilumod_OutputStringVec		},
    {	"InputStringVec",		ilumod_InputStringVec		},

    {	"SizeOfWStringVec",		ilumod_SizeOfWStringVec		},
    {	"OutputWStringVec",		ilumod_OutputWStringVec		},
    {	"InputWStringVec",		ilumod_InputWStringVec		},

    {	"SizeOfOptional",		ilumod_SizeOfOptional		},
    {	"OutputOptional",		ilumod_OutputOptional		},
    {	"InputOptional",		ilumod_InputOptional		},

    {	"SizeOfUnion",			ilumod_SizeOfUnion		},
    {	"OutputUnion",			ilumod_OutputUnion		},
    {	"InputUnion",			ilumod_InputUnion		},
    {	"EndUnion",			ilumod_EndUnion			},

    {	"SizeOfArray",			ilumod_SizeOfArray		},
    {	"OutputArray",			ilumod_OutputArray		},
    {	"InputArray",			ilumod_InputArray		},
    {	"EndArray",			ilumod_EndArray			},

    {	"SizeOfRecord",			ilumod_SizeOfRecord		},
    {	"OutputRecord",			ilumod_OutputRecord		},
    {	"InputRecord",			ilumod_InputRecord		},
    {	"EndRecord",			ilumod_EndRecord		},

    {	"SizeOfSequence",		ilumod_SizeOfSequence		},
    {	"OutputSequence",		ilumod_OutputSequence		},
    {	"InputSequence",		ilumod_InputSequence		},
    {	"EndSequence",			ilumod_EndSequence		},

    {	0								}
};

/********************************/

static PyObject *
newException(PyObject *dict, char *name)
{
	PyObject *	e	= PyString_FromString(name);

	if (e == 0 || PyDict_SetItemString(dict, name, e) < 0)
		Py_FatalError("ilu can't define exception");
	return e;
}

static void
createExceptions(PyObject *modDict)
{
	GeneralError = newException(modDict, "IluGeneralError");
	ProtocolError = newException(modDict, "IluProtocolError");
	UnimplementedMethodError = newException(modDict,
		"IluUnimplementedMethodError");
}

/********************************/

static void
createConstants(PyObject *modDict)
{
	PyObject *	FineTimeRate;
	PyObject *	Version;

	if (PyDict_SetItemString(modDict, "FALSE", Py_False) < 0 ||
	    PyDict_SetItemString(modDict, "TRUE", Py_True) < 0)
		Py_FatalError("ilu can't define FALSE and TRUE");

	if ((FineTimeRate = PyInt_FromLong(ilu_FineTimeRate)) == 0 ||
	    PyDict_SetItemString(modDict, "FineTimeRate", FineTimeRate) < 0)
		Py_FatalError("ilu can't define FineTimeRate");

	if ((Version = PyString_FromString(ilu_GetILUVersion())) == 0 ||
	    PyDict_SetItemString(modDict, "Version", Version) < 0)
		Py_FatalError("ilu can't define Version");
}

/********************************/

void
initiluPr(void)
{
	PyObject *	mod	= Py_InitModule("iluPr", ilumod_methods);
	PyObject *	dict	= PyModule_GetDict(mod);

	createExceptions(dict);
	createConstants(dict);

	ilu_SetNoter (trackKernelInterest);
}
