/*
Copyright (c) 1991, 1992, 1993, 1994 Xerox Corporation.  All Rights Reserved.

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: iluclobject.c,v 1.10 1994/12/09 18:36:56 severson Exp $
*/

/* from ILU */
#include "iluxport.h"

/* local */
#include "python.h"
#include "iluclobject.h"

static void
ilucl_deallocSkeletons(IluclObject *cl)
{
	if (cl->skeletons)
	{
		int	i;

		for (i = 0; i < cl->c.cl_method_count; i++)
			Py_XDECREF(cl->skeletons[i]);
	}
	PyMem_XDEL(cl->skeletons);
}

static void
ilucl_dealloc(PyObject *o)
{
	IluclObject *	cl = (IluclObject *) o;
	int		i;

	PyMem_XDEL(cl->c.cl_name);
	PyMem_XDEL(cl->c.cl_brand);
	PyMem_XDEL(cl->c.cl_name);
	PyMem_XDEL(cl->c.cl_unique_id);
	PyMem_XDEL(cl->c.cl_singleton);
	PyMem_XDEL(cl->c.cl_authentication);

	for (i = 0; i < cl->c.cl_method_count; i++)
	{
		ilu_Method	m	= cl->c.cl_methods + i;
		int		j;

		for (j = 0; j < m->me_exceptionCount; j++)
			Py_XDECREF(cl->excepts[i][j]);
		PyMem_XDEL(cl->excepts[i]);
		PyMem_XDEL(m->me_name);
		PyMem_XDEL(m->me_exceptionVector);
	}
	PyMem_XDEL(cl->c.cl_methods);

	for (i = 0; i < cl->c.cl_scls_count; i++)
		Py_XDECREF(cl->superclasses[i]);
	PyMem_XDEL(cl->superclasses);
	PyMem_XDEL(cl->c.cl_scls_ids);
	PyMem_XDEL(cl->c.cl_sclses);

	ilucl_deallocSkeletons(cl);

	PyMem_DEL(o);
}

#define CLEAR(ptr, n)	memset((char *) ptr, 0, (n) * sizeof(*(ptr)))

static int
setMethodEntry(ilu_Method m, PyObject ***pExcObjs, PyObject *t)
{
	char *		name;
	int		id;
	char		cacheable;
	char		asynchronous;
	PyObject *	exceptionsTuple;
	int		i;

	if (!PyArg_Parse(t, "(sibbO)", &name, &id, &cacheable, &asynchronous,
	    &exceptionsTuple))
		return -1;
	if ((m->me_name = malloc(strlen(name) + 1)) == 0)
	{
		(void) PyErr_NoMemory();
		return -1;
	}
	strcpy(m->me_name, name);
	m->me_id = id;
	m->me_cacheable = cacheable;
	m->me_asynchronous = asynchronous;

	m->me_exceptionCount = PyTuple_Size(exceptionsTuple);
	m->me_exceptionVector = PyMem_NEW(char *, m->me_exceptionCount);
	if (m->me_exceptionVector == 0)
	{
		PyMem_XDEL(m->me_name);
		m->me_exceptionCount = 0;
		(void) PyErr_NoMemory();
		return -1;
	}
	*pExcObjs = PyMem_NEW(PyObject *, m->me_exceptionCount);
	if (*pExcObjs == 0)
	{
		PyMem_XDEL(m->me_name);
		PyMem_XDEL(m->me_exceptionVector);
		m->me_exceptionCount = 0;
		(void) PyErr_NoMemory();
		return -1;
	}
	for (i = 0; i < m->me_exceptionCount; i++)
	{
		PyObject *	e = PyTuple_GetItem(exceptionsTuple, i);

		if (PyString_Check(e))
		{
			Py_INCREF(e);
			(*pExcObjs)[i] = e;
			m->me_exceptionVector[i] = PyString_AsString(e);
		}
		else
		{
			PyMem_XDEL(m->me_name);
			PyMem_XDEL(m->me_exceptionVector);
			PyMem_XDEL(*pExcObjs);
			(void) PyErr_BadArgument();
			return -1;
		}
	}

	return 0;
}

PyObject *
ilucl_New(char *name, char *brand, char *uniqueId, char *singleton,
		int collectible, int optional, char *authentication,
		PyObject *methodsTuple, PyObject *superclassTuple)
{
	IluclObject *	cl	= PyObject_NEW(IluclObject, &Ilucl_Type);
	int		i;

	if (cl == 0)
		return 0;

	/* Clear fields before possible dealloc. */
	CLEAR(&cl->c, 1);
	cl->excepts = 0;
	cl->superclasses = 0;
	cl->skeletons = 0;

	cl->c.cl_name = malloc(strlen(name) + 1);
	cl->c.cl_brand = malloc(strlen(brand) + 1);
	cl->c.cl_unique_id = malloc(strlen(uniqueId) + 1);
	if (cl->c.cl_name == 0 || cl->c.cl_brand == 0 ||
	    cl->c.cl_unique_id == 0)
	{
		ilucl_dealloc((PyObject *) cl);
		return PyErr_NoMemory();
	}
	strcpy(cl->c.cl_name, name);
	strcpy(cl->c.cl_brand, brand);
	strcpy(cl->c.cl_unique_id, uniqueId);


	if (singleton)
	{
		cl->c.cl_singleton = malloc(strlen(singleton) + 1);
		if (cl->c.cl_singleton == 0)
		{
			ilucl_dealloc((PyObject *) cl);
			return PyErr_NoMemory();
		}
		strcpy(cl->c.cl_singleton, singleton);
	}

	cl->c.cl_collectible = collectible;
	cl->c.cl_optional = optional;

	if (authentication)
	{
		cl->c.cl_authentication = malloc(strlen(authentication) + 1);
		if (cl->c.cl_authentication == 0)
		{
			ilucl_dealloc((PyObject *) cl);
			return PyErr_NoMemory();
		}
		strcpy(cl->c.cl_authentication, authentication);
	}

	cl->c.cl_method_count = PyTuple_Size(methodsTuple);
	cl->c.cl_methods = PyMem_NEW(struct _ilu_Method_s,
					cl->c.cl_method_count);
	if (cl->c.cl_methods == 0)
	{
		cl->c.cl_method_count = 0;
		ilucl_dealloc((PyObject *) cl);
		return PyErr_NoMemory();
	}
	CLEAR(cl->c.cl_methods, cl->c.cl_method_count);

	cl->excepts = PyMem_NEW(PyObject **, cl->c.cl_method_count);
	if (cl->excepts == 0)
	{
		PyMem_XDEL(cl->c.cl_methods);
		cl->c.cl_method_count = 0;
		ilucl_dealloc((PyObject *) cl);
		return PyErr_NoMemory();
	}
	CLEAR(cl->excepts, cl->c.cl_method_count);

	for (i = 0; i < cl->c.cl_method_count; i++)
	{
		if (setMethodEntry(cl->c.cl_methods + i, cl->excepts + i,
		    PyTuple_GetItem(methodsTuple, i)) < 0)
		{
			cl->c.cl_method_count = i;
			ilucl_dealloc((PyObject *) cl);
			return 0;
		}
	}

	cl->c.cl_scls_count = PyTuple_Size(superclassTuple);
	cl->superclasses = PyMem_NEW(PyObject *, cl->c.cl_scls_count);
	cl->c.cl_scls_ids = PyMem_NEW(char *, cl->c.cl_scls_count);
	cl->c.cl_sclses = PyMem_NEW(ilu_Class, cl->c.cl_scls_count);
	if (cl->superclasses == 0 || cl->c.cl_scls_ids == 0 ||
	    cl->c.cl_sclses == 0)
	{
		cl->c.cl_scls_count = 0;
		PyMem_XDEL(cl->superclasses);
		PyMem_XDEL(cl->c.cl_scls_ids);
		PyMem_XDEL(cl->c.cl_sclses);
		ilucl_dealloc((PyObject *) cl);
		return 0;
	}
	for (i = 0; i < cl->c.cl_scls_count; i++)
	{
		IluclObject *	sc	=
			(IluclObject *) PyTuple_GetItem(superclassTuple, i);

		if (!ilucl_Check(sc))
		{
			cl->c.cl_scls_count = i;
			ilucl_dealloc((PyObject *) cl);
			(void) PyErr_BadArgument();
			return 0;
		}
		Py_INCREF(sc);
		cl->superclasses[i] = (PyObject *) sc;
		cl->c.cl_scls_ids[i] = sc->c.cl_unique_id;
		cl->c.cl_sclses[i] = &sc->c;
	}

	return (PyObject *) cl;
}

int
ilucl_RegisterSkeletons(IluclObject *cl, PyObject *skelTuple)
{
	int	i;

	if (PyTuple_Size(skelTuple) != cl->c.cl_method_count)
	{
		PyErr_SetString(PyExc_TypeError, "skel tuple wrong length");
		return -1;
	}

	ilucl_deallocSkeletons(cl);
	cl->skeletons = PyMem_NEW(PyObject *, cl->c.cl_method_count);
	if (cl->skeletons == 0)
	{
		(void) PyErr_NoMemory();
		return -1;
	}
	CLEAR(cl->skeletons, cl->c.cl_method_count);

	for (i = 0; i < cl->c.cl_method_count; i++)
	{
		PyObject *	skel	 = PyTuple_GetItem(skelTuple, i);

		if (!PyCallable_Check(skel))
		{
			ilucl_deallocSkeletons(cl);
			PyErr_SetString(PyExc_TypeError,
				"skel tuple contains noncallable");
			return -1;
		}
		Py_INCREF(skel);
		cl->skeletons[i] = skel;
	}

	return 0;
}

static char *
booleanImage(int bool)
{
	return bool ? "TRUE" : "FALSE";
}

static char *
notPrefix(int bool)
{
	return bool ? "" : "not ";
}

static int
ilucl_print(PyObject *o, FILE *fp, int flags)
{
	IluclObject *	cl = (IluclObject *) o;
	int		i;

	fprintf(fp, "<ilu_Class at 0x%p\n", cl);
	fprintf(fp, "  name: '%s'\n", cl->c.cl_name);
	fprintf(fp, "  brand: '%s'\n", cl->c.cl_brand);
	fprintf(fp, "  unique_id: '%s'\n", cl->c.cl_unique_id);

	fprintf(fp, "  singleton: ");
	if (cl->c.cl_singleton)
		fprintf(fp, "'%s'", cl->c.cl_singleton);
	else
		fprintf(fp, "null");
	putc('\n', fp);

	fprintf(fp, "  collectible: %s\n", booleanImage(cl->c.cl_collectible));
	fprintf(fp, "  optional: %s\n", booleanImage(cl->c.cl_optional));

	fprintf(fp, "  authentication: ");
	if (cl->c.cl_authentication)
		fprintf(fp, "'%s'", cl->c.cl_authentication);
	else
		fprintf(fp, "null");
	putc('\n', fp);

	for (i = 0; i < cl->c.cl_method_count; i++)
	{
		ilu_Method	m	= cl->c.cl_methods + i;

		if (i == 0)
			fprintf(fp, "  methods:\n");
		fprintf(fp, "    '%s', %lu, %scacheable, %sasync\n",
			m->me_name, (unsigned long) m->me_id,
			notPrefix(m->me_cacheable),
			notPrefix(m->me_asynchronous));
		if (m->me_exceptionCount > 0)
		{
			int	i;

			fprintf(fp, "      raises ");
			for (i = 0; i < m->me_exceptionCount; i++)
			{
				fprintf(fp, "%s'%s'", i == 0 ? "" : ", ",
					m->me_exceptionVector[i]);
			}
			putc('\n', fp);
		}
	}

	for (i = 0; i < cl->c.cl_scls_count; i++)
	{
		if (i == 0)
			fprintf(fp, "  superclass ids:\n");
		fprintf(fp, "    '%s'\n", cl->c.cl_scls_ids[i]);
	}

	for (i = 0; i < cl->c.cl_scls_count; i++)
	{
		if (i == 0)
			fprintf(fp, "  superclasses:\n");
		fprintf(fp, "    %s\n", cl->c.cl_sclses[i]->cl_name);
	}

	if (cl->skeletons)
	{
		fprintf(fp, "  skeletons:\n");
		for (i = 0; i < cl->c.cl_method_count; i++)
		{
			fprintf(fp, "    ");
			PyObject_Print(cl->skeletons[i], fp, flags);
			fprintf(fp, "\n");
		}
	}

	fprintf(fp, "  >");
	return 0;
}

static PyObject *
ilucl_name(PyObject *self, PyObject *args)
{
	IluclObject *	cl	= (IluclObject *) self;

	if (!PyArg_Parse(args, ""))
		return 0;
	return PyString_FromString(cl->c.cl_name);
}

static PyObject *
ilucl_id(PyObject *self, PyObject *args)
{
	IluclObject *	cl	= (IluclObject *) self;

	if (!PyArg_Parse(args, ""))
		return 0;
	return PyString_FromString(cl->c.cl_unique_id);
}

static PyMethodDef ilucl_methods[] =
{
	{ "name",		ilucl_name		},
	{ "id",			ilucl_id		},
	{ 0						}
};

static PyObject *
ilucl_getattr(PyObject *self, char *name)
{
	return Py_FindMethod(ilucl_methods, self, name);
}

PyTypeObject	Ilucl_Type =
{
	PyObject_HEAD_INIT(&PyType_Type)
	0,
	"ilu_Class",
	sizeof(IluclObject),
	0,
	ilucl_dealloc,		/*tp_dealloc*/
	ilucl_print,		/*tp_print*/
	ilucl_getattr,		/*tp_getattr*/
	0,			/*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
};
